import asyncio
import pytest

from aioredis import ReplyError, MultiExecError, WatchVariableError
from aioredis import ConnectionClosedError


@pytest.mark.run_loop
async def test_multi_exec(redis, loop):
    await redis.delete('foo', 'bar')

    tr = redis.multi_exec()
    f1 = tr.incr('foo')
    f2 = tr.incr('bar')
    res = await tr.execute()
    assert res == [1, 1]
    res2 = await asyncio.gather(f1, f2, loop=loop)
    assert res == res2

    tr = redis.multi_exec()
    f1 = tr.incr('foo')
    f2 = tr.incr('bar')
    await tr.execute()
    assert (await f1) == 2
    assert (await f2) == 2

    tr = redis.multi_exec()
    f1 = tr.set('foo', 1.0)
    f2 = tr.incrbyfloat('foo', 1.2)
    res = await tr.execute()
    assert res == [True, 2.2]
    res2 = await asyncio.gather(f1, f2, loop=loop)
    assert res == res2

    tr = redis.multi_exec()
    f1 = tr.incrby('foo', 1.0)
    with pytest.raises(MultiExecError, match="increment must be .* int"):
        await tr.execute()
    with pytest.raises(TypeError):
        await f1


@pytest.mark.run_loop
async def test_empty(redis):
    tr = redis.multi_exec()
    res = await tr.execute()
    assert res == []


@pytest.mark.run_loop
async def test_double_execute(redis):
    tr = redis.multi_exec()
    await tr.execute()
    with pytest.raises(AssertionError):
        await tr.execute()
    with pytest.raises(AssertionError):
        await tr.incr('foo')


@pytest.mark.run_loop
async def test_connection_closed(redis):
    tr = redis.multi_exec()
    fut1 = tr.quit()
    fut2 = tr.incrby('foo', 1.0)
    fut3 = tr.incrby('foo', 1)
    with pytest.raises(MultiExecError):
        await tr.execute()

    assert fut1.done() is True
    assert fut2.done() is True
    assert fut3.done() is True
    assert fut1.exception() is not None
    assert fut2.exception() is not None
    assert fut3.exception() is not None
    assert not fut1.cancelled()
    assert not fut2.cancelled()
    assert not fut3.cancelled()

    try:
        assert (await fut1) == b'OK'
    except Exception as err:
        assert isinstance(err, (ConnectionClosedError, ConnectionError))
    assert fut2.cancelled() is False
    assert isinstance(fut2.exception(), TypeError)

    # assert fut3.cancelled() is True
    assert fut3.done() and not fut3.cancelled()
    assert isinstance(fut3.exception(),
                      (ConnectionClosedError, ConnectionError))


@pytest.mark.run_loop
async def test_discard(redis):
    await redis.delete('foo')
    tr = redis.multi_exec()
    fut1 = tr.incrby('foo', 1.0)
    fut2 = tr.connection.execute('MULTI')
    fut3 = tr.connection.execute('incr', 'foo')

    with pytest.raises(MultiExecError):
        await tr.execute()
    with pytest.raises(TypeError):
        await fut1
    with pytest.raises(ReplyError):
        await fut2
    # with pytest.raises(ReplyError):
    res = await fut3
    assert res == 1


@pytest.mark.run_loop
async def test_exec_error(redis):
    tr = redis.multi_exec()
    fut = tr.connection.execute('INCRBY', 'key', '1.0')
    with pytest.raises(MultiExecError):
        await tr.execute()
    with pytest.raises(ReplyError):
        await fut

    await redis.set('foo', 'bar')
    tr = redis.multi_exec()
    fut = tr.incrbyfloat('foo', 1.1)
    res = await tr.execute(return_exceptions=True)
    assert isinstance(res[0], ReplyError)
    with pytest.raises(ReplyError):
        await fut


@pytest.mark.run_loop
async def test_command_errors(redis):
    tr = redis.multi_exec()
    fut = tr.incrby('key', 1.0)
    with pytest.raises(MultiExecError):
        await tr.execute()
    with pytest.raises(TypeError):
        await fut


@pytest.mark.run_loop
async def test_several_command_errors(redis):
    tr = redis.multi_exec()
    fut1 = tr.incrby('key', 1.0)
    fut2 = tr.rename('bar', 'bar')
    with pytest.raises(MultiExecError):
        await tr.execute()
    with pytest.raises(TypeError):
        await fut1
    with pytest.raises(ValueError):
        await fut2


@pytest.mark.run_loop
async def test_error_in_connection(redis):
    await redis.set('foo', 1)
    tr = redis.multi_exec()
    fut1 = tr.mget('foo', None)
    fut2 = tr.incr('foo')
    with pytest.raises(MultiExecError):
        await tr.execute()
    with pytest.raises(TypeError):
        await fut1
    await fut2


@pytest.mark.run_loop
async def test_watch_unwatch(redis):
    res = await redis.watch('key')
    assert res is True
    res = await redis.watch('key', 'key')
    assert res is True

    with pytest.raises(TypeError):
        await redis.watch(None)
    with pytest.raises(TypeError):
        await redis.watch('key', None)
    with pytest.raises(TypeError):
        await redis.watch('key', 'key', None)

    res = await redis.unwatch()
    assert res is True


@pytest.mark.run_loop
async def test_encoding(redis):
    res = await redis.set('key', 'value')
    assert res is True
    res = await redis.hmset(
        'hash-key', 'foo', 'val1', 'bar', 'val2')
    assert res is True

    tr = redis.multi_exec()
    fut1 = tr.get('key')
    fut2 = tr.get('key', encoding='utf-8')
    fut3 = tr.hgetall('hash-key', encoding='utf-8')
    await tr.execute()
    res = await fut1
    assert res == b'value'
    res = await fut2
    assert res == 'value'
    res = await fut3
    assert res == {'foo': 'val1', 'bar': 'val2'}


@pytest.mark.run_loop
async def test_global_encoding(redis, create_redis, server, loop):
    redis = await create_redis(
        server.tcp_address,
        loop=loop, encoding='utf-8')
    res = await redis.set('key', 'value')
    assert res is True
    res = await redis.hmset(
        'hash-key', 'foo', 'val1', 'bar', 'val2')
    assert res is True

    tr = redis.multi_exec()
    fut1 = tr.get('key')
    fut2 = tr.get('key', encoding='utf-8')
    fut3 = tr.hgetall('hash-key', encoding='utf-8')
    await tr.execute()
    res = await fut1
    assert res == 'value'
    res = await fut2
    assert res == 'value'
    res = await fut3
    assert res == {'foo': 'val1', 'bar': 'val2'}


@pytest.mark.run_loop
async def test_transaction__watch_error(redis, create_redis, server, loop):
    other = await create_redis(
        server.tcp_address, loop=loop)

    ok = await redis.set('foo', 'bar')
    assert ok is True

    ok = await redis.watch('foo')
    assert ok is True

    ok = await other.set('foo', 'baz')
    assert ok is True

    tr = redis.multi_exec()
    fut1 = tr.set('foo', 'foo')
    fut2 = tr.get('bar')
    with pytest.raises(MultiExecError):
        await tr.execute()
    with pytest.raises(WatchVariableError):
        await fut1
    with pytest.raises(WatchVariableError):
        await fut2


@pytest.mark.run_loop
async def test_multi_exec_and_pool_release(redis):
    # Test the case when pool connection is released before
    # `exec` result is received.

    slow_script = """
    local a = tonumber(redis.call('time')[1])
    local b = a + 1
    while (a < b)
    do
        a = tonumber(redis.call('time')[1])
    end
    """

    tr = redis.multi_exec()
    fut1 = tr.eval(slow_script)
    ret, = await tr.execute()
    assert ret is None
    assert (await fut1) is None


@pytest.mark.run_loop
async def test_multi_exec_db_select(redis):
    await redis.set('foo', 'bar')

    tr = redis.multi_exec()
    f1 = tr.get('foo', encoding='utf-8')
    f2 = tr.get('foo')
    await tr.execute()
    assert await f1 == 'bar'
    assert await f2 == b'bar'



import asyncio
import time
import math
import pytest
import sys

from unittest import mock

from aioredis import ReplyError


async def add(redis, key, value):
    ok = await redis.connection.execute('set', key, value)
    assert ok == b'OK'


@pytest.mark.run_loop
async def test_delete(redis):
    await add(redis, 'my-key', 123)
    await add(redis, 'other-key', 123)

    res = await redis.delete('my-key', 'non-existent-key')
    assert res == 1

    res = await redis.delete('other-key', 'other-key')
    assert res == 1

    with pytest.raises(TypeError):
        await redis.delete(None)

    with pytest.raises(TypeError):
        await redis.delete('my-key', 'my-key', None)


@pytest.mark.run_loop
async def test_dump(redis):
    await add(redis, 'my-key', 123)

    data = await redis.dump('my-key')
    assert data == mock.ANY
    assert isinstance(data, (bytes, bytearray))
    assert len(data) > 0

    data = await redis.dump('non-existent-key')
    assert data is None

    with pytest.raises(TypeError):
        await redis.dump(None)


@pytest.mark.run_loop
async def test_exists(redis, server):
    await add(redis, 'my-key', 123)

    res = await redis.exists('my-key')
    assert isinstance(res, int)
    assert res == 1

    res = await redis.exists('non-existent-key')
    assert isinstance(res, int)
    assert res == 0

    with pytest.raises(TypeError):
        await redis.exists(None)
    if server.version < (3, 0, 3):
        with pytest.raises(ReplyError):
            await redis.exists('key-1', 'key-2')


@pytest.redis_version(
    3, 0, 3, reason='Multi-key EXISTS available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_exists_multiple(redis):
    await add(redis, 'my-key', 123)

    res = await redis.exists('my-key', 'other-key')
    assert isinstance(res, int)
    assert res == 1

    res = await redis.exists('my-key', 'my-key')
    assert isinstance(res, int)
    assert res == 2

    res = await redis.exists('foo', 'bar')
    assert isinstance(res, int)
    assert res == 0


@pytest.mark.run_loop
async def test_expire(redis):
    await add(redis, 'my-key', 132)

    res = await redis.expire('my-key', 10)
    assert res is True

    res = await redis.connection.execute('TTL', 'my-key')
    assert res >= 10

    await redis.expire('my-key', -1)
    res = await redis.exists('my-key')
    assert not res

    res = await redis.expire('other-key', 1000)
    assert res is False

    await add(redis, 'my-key', 1)
    res = await redis.expire('my-key', 10.0)
    assert res is True
    res = await redis.connection.execute('TTL', 'my-key')
    assert res >= 10

    with pytest.raises(TypeError):
        await redis.expire(None, 123)
    with pytest.raises(TypeError):
        await redis.expire('my-key', 'timeout')


@pytest.mark.run_loop
async def test_expireat(redis):
    await add(redis, 'my-key', 123)
    now = math.ceil(time.time())

    fut1 = redis.expireat('my-key', now + 10)
    fut2 = redis.connection.execute('TTL', 'my-key')
    assert (await fut1) is True
    assert (await fut2) >= 10

    now = time.time()
    fut1 = redis.expireat('my-key', now + 10)
    fut2 = redis.connection.execute('TTL', 'my-key')
    assert (await fut1) is True
    assert (await fut2) >= 10

    res = await redis.expireat('my-key', -1)
    assert res is True

    res = await redis.exists('my-key')
    assert not res

    await add(redis, 'my-key', 123)

    res = await redis.expireat('my-key', 0)
    assert res is True

    res = await redis.exists('my-key')
    assert not res

    await add(redis, 'my-key', 123)
    with pytest.raises(TypeError):
        await redis.expireat(None, 123)
    with pytest.raises(TypeError):
        await redis.expireat('my-key', 'timestamp')


@pytest.mark.run_loop
async def test_keys(redis):
    res = await redis.keys('*pattern*')
    assert res == []

    await redis.connection.execute('FLUSHDB')
    res = await redis.keys('*')
    assert res == []

    await add(redis, 'my-key-1', 1)
    await add(redis, 'my-key-ab', 1)

    res = await redis.keys('my-key-?')
    assert res == [b'my-key-1']
    res = await redis.keys('my-key-*')
    assert sorted(res) == [b'my-key-1', b'my-key-ab']

    # test with encoding param
    res = await redis.keys('my-key-*', encoding='utf-8')
    assert sorted(res) == ['my-key-1', 'my-key-ab']

    with pytest.raises(TypeError):
        await redis.keys(None)


@pytest.mark.run_loop
async def test_migrate(create_redis, loop, server, serverB):
    redisA = await create_redis(server.tcp_address)
    redisB = await create_redis(serverB.tcp_address, db=2)

    await add(redisA, 'my-key', 123)

    await redisB.delete('my-key')
    assert (await redisA.exists('my-key'))
    assert not (await redisB.exists('my-key'))

    ok = await redisA.migrate(
        'localhost', serverB.tcp_address.port, 'my-key', 2, 1000)
    assert ok is True
    assert not (await redisA.exists('my-key'))
    assert (await redisB.exists('my-key'))

    with pytest.raises(TypeError, match="host .* str"):
        await redisA.migrate(None, 1234, 'key', 1, 23)
    with pytest.raises(TypeError, match="args .* None"):
        await redisA.migrate('host', '1234',  None, 1, 123)
    with pytest.raises(TypeError, match="dest_db .* int"):
        await redisA.migrate('host', 123, 'key', 1.0, 123)
    with pytest.raises(TypeError, match="timeout .* int"):
        await redisA.migrate('host', '1234', 'key', 2, None)
    with pytest.raises(ValueError, match="Got empty host"):
        await redisA.migrate('', '123', 'key', 1, 123)
    with pytest.raises(ValueError, match="dest_db .* greater equal 0"):
        await redisA.migrate('host', 6379, 'key', -1, 1000)
    with pytest.raises(ValueError, match="timeout .* greater equal 0"):
        await redisA.migrate('host', 6379, 'key', 1, -1000)


@pytest.redis_version(
    3, 0, 0, reason="Copy/Replace flags available since Redis 3.0")
@pytest.mark.run_loop
async def test_migrate_copy_replace(create_redis, loop, server, serverB):
    redisA = await create_redis(server.tcp_address)
    redisB = await create_redis(serverB.tcp_address, db=0)

    await add(redisA, 'my-key', 123)
    await redisB.delete('my-key')

    ok = await redisA.migrate(
        'localhost', serverB.tcp_address.port, 'my-key', 0, 1000, copy=True)
    assert ok is True
    assert (await redisA.get('my-key')) == b'123'
    assert (await redisB.get('my-key')) == b'123'

    assert (await redisA.set('my-key', 'val'))
    ok = await redisA.migrate(
        'localhost', serverB.tcp_address.port, 'my-key', 2, 1000, replace=True)
    assert (await redisA.get('my-key')) is None
    assert (await redisB.get('my-key'))


@pytest.redis_version(
    3, 0, 6, reason="MIGRATE…KEYS available since Redis 3.0.6")
@pytest.mark.skipif(
    sys.platform == 'win32', reason="Seems to be unavailable in win32 build")
@pytest.mark.run_loop
async def test_migrate_keys(create_redis, loop, server, serverB):
    redisA = await create_redis(server.tcp_address)
    redisB = await create_redis(serverB.tcp_address, db=0)

    await add(redisA, 'key1', 123)
    await add(redisA, 'key2', 123)
    await add(redisA, 'key3', 123)
    await redisB.delete('key1', 'key2', 'key3')

    ok = await redisA.migrate_keys(
        'localhost', serverB.tcp_address.port,
        ('key1', 'key2', 'key3', 'non-existing-key'),
        dest_db=0, timeout=1000)
    assert ok is True

    assert (await redisB.get('key1')) == b'123'
    assert (await redisB.get('key2')) == b'123'
    assert (await redisB.get('key3')) == b'123'
    assert (await redisA.get('key1')) is None
    assert (await redisA.get('key2')) is None
    assert (await redisA.get('key3')) is None

    ok = await redisA.migrate_keys(
        'localhost', serverB.tcp_address.port, ('key1', 'key2', 'key3'),
        dest_db=0, timeout=1000)
    assert not ok
    ok = await redisB.migrate_keys(
        'localhost', server.tcp_address.port, ('key1', 'key2', 'key3'),
        dest_db=0, timeout=1000,
        copy=True)
    assert ok
    assert (await redisB.get('key1')) == b'123'
    assert (await redisB.get('key2')) == b'123'
    assert (await redisB.get('key3')) == b'123'
    assert (await redisA.get('key1')) == b'123'
    assert (await redisA.get('key2')) == b'123'
    assert (await redisA.get('key3')) == b'123'

    assert (await redisA.set('key1', 'val'))
    assert (await redisA.set('key2', 'val'))
    assert (await redisA.set('key3', 'val'))
    ok = await redisA.migrate_keys(
        'localhost', serverB.tcp_address.port,
        ('key1', 'key2', 'key3', 'non-existing-key'),
        dest_db=0, timeout=1000, replace=True)
    assert ok is True

    assert (await redisB.get('key1')) == b'val'
    assert (await redisB.get('key2')) == b'val'
    assert (await redisB.get('key3')) == b'val'
    assert (await redisA.get('key1')) is None
    assert (await redisA.get('key2')) is None
    assert (await redisA.get('key3')) is None


@pytest.mark.run_loop
async def test_migrate__exceptions(redis, loop, server, unused_port):
    await add(redis, 'my-key', 123)

    assert (await redis.exists('my-key'))

    with pytest.raises(ReplyError, match="IOERR .* timeout .*"):
        assert not (await redis.migrate(
            'localhost', unused_port(),
            'my-key', dest_db=30, timeout=10))


@pytest.redis_version(
    3, 0, 6, reason="MIGRATE…KEYS available since Redis 3.0.6")
@pytest.mark.skipif(
    sys.platform == 'win32', reason="Seems to be unavailable in win32 build")
@pytest.mark.run_loop
async def test_migrate_keys__errors(redis):
    with pytest.raises(TypeError, match="host .* str"):
        await redis.migrate_keys(None, 1234, 'key', 1, 23)
    with pytest.raises(TypeError, match="keys .* list or tuple"):
        await redis.migrate_keys('host', '1234',  None, 1, 123)
    with pytest.raises(TypeError, match="dest_db .* int"):
        await redis.migrate_keys('host', 123, ('key',), 1.0, 123)
    with pytest.raises(TypeError, match="timeout .* int"):
        await redis.migrate_keys('host', '1234', ('key',), 2, None)
    with pytest.raises(ValueError, match="Got empty host"):
        await redis.migrate_keys('', '123', ('key',), 1, 123)
    with pytest.raises(ValueError, match="dest_db .* greater equal 0"):
        await redis.migrate_keys('host', 6379, ('key',), -1, 1000)
    with pytest.raises(ValueError, match="timeout .* greater equal 0"):
        await redis.migrate_keys('host', 6379, ('key',), 1, -1000)
    with pytest.raises(ValueError, match="keys .* empty"):
        await redis.migrate_keys('host', '1234', (), 2, 123)


@pytest.mark.run_loop
async def test_move(redis):
    await add(redis, 'my-key', 123)

    assert redis.db == 0
    res = await redis.move('my-key', 1)
    assert res is True

    with pytest.raises(TypeError):
        await redis.move(None, 1)
    with pytest.raises(TypeError):
        await redis.move('my-key', None)
    with pytest.raises(ValueError):
        await redis.move('my-key', -1)
    with pytest.raises(TypeError):
        await redis.move('my-key', 'not db')


@pytest.mark.run_loop
async def test_object_refcount(redis):
    await add(redis, 'foo', 'bar')

    res = await redis.object_refcount('foo')
    assert res == 1
    res = await redis.object_refcount('non-existent-key')
    assert res is None

    with pytest.raises(TypeError):
        await redis.object_refcount(None)


@pytest.mark.run_loop
async def test_object_encoding(redis, server):
    await add(redis, 'foo', 'bar')

    res = await redis.object_encoding('foo')

    if server.version < (3, 0, 0):
        assert res == b'raw'
    else:
        assert res == b'embstr'

    res = await redis.incr('key')
    assert res == 1
    res = await redis.object_encoding('key')
    assert res == b'int'
    res = await redis.object_encoding('non-existent-key')
    assert res is None

    with pytest.raises(TypeError):
        await redis.object_encoding(None)


@pytest.mark.run_loop(timeout=20)
async def test_object_idletime(redis, loop, server):
    await add(redis, 'foo', 'bar')

    res = await redis.object_idletime('foo')
    # NOTE: sometimes travis-ci is too slow
    assert res >= 0

    res = 0
    while not res:
        res = await redis.object_idletime('foo')
        await asyncio.sleep(.5, loop=loop)
    assert res >= 1

    res = await redis.object_idletime('non-existent-key')
    assert res is None

    with pytest.raises(TypeError):
        await redis.object_idletime(None)


@pytest.mark.run_loop
async def test_persist(redis):
    await add(redis, 'my-key', 123)
    res = await redis.expire('my-key', 10)
    assert res is True

    res = await redis.persist('my-key')
    assert res is True

    res = await redis.connection.execute('TTL', 'my-key')
    assert res == -1

    with pytest.raises(TypeError):
        await redis.persist(None)


@pytest.mark.run_loop
async def test_pexpire(redis, loop):
    await add(redis, 'my-key', 123)
    res = await redis.pexpire('my-key', 100)
    assert res is True

    res = await redis.connection.execute('TTL', 'my-key')
    assert res == 0
    res = await redis.connection.execute('PTTL', 'my-key')
    assert res > 0

    await add(redis, 'my-key', 123)
    res = await redis.pexpire('my-key', 1)
    assert res is True

    # XXX: tests now looks strange to me.
    await asyncio.sleep(.2, loop=loop)

    res = await redis.exists('my-key')
    assert not res

    with pytest.raises(TypeError):
        await redis.pexpire(None, 0)
    with pytest.raises(TypeError):
        await redis.pexpire('my-key', 1.0)


@pytest.mark.run_loop
async def test_pexpireat(redis):
    await add(redis, 'my-key', 123)
    now = math.ceil((await redis.time()) * 1000)
    fut1 = redis.pexpireat('my-key', now + 2000)
    fut2 = redis.ttl('my-key')
    fut3 = redis.pttl('my-key')
    assert (await fut1) is True
    assert (await fut2) == 2
    pytest.assert_almost_equal((await fut3), 2000, -3)

    with pytest.raises(TypeError):
        await redis.pexpireat(None, 1234)
    with pytest.raises(TypeError):
        await redis.pexpireat('key', 'timestamp')
    with pytest.raises(TypeError):
        await redis.pexpireat('key', 1000.0)


@pytest.mark.run_loop
async def test_pttl(redis, server):
    await add(redis, 'key', 'val')
    res = await redis.pttl('key')
    assert res == -1
    res = await redis.pttl('non-existent-key')
    if server.version < (2, 8, 0):
        assert res == -1
    else:
        assert res == -2

    await redis.pexpire('key', 500)
    res = await redis.pttl('key')
    pytest.assert_almost_equal(res, 500, -2)

    with pytest.raises(TypeError):
        await redis.pttl(None)


@pytest.mark.run_loop
async def test_randomkey(redis):
    await add(redis, 'key:1', 123)
    await add(redis, 'key:2', 123)
    await add(redis, 'key:3', 123)

    res = await redis.randomkey()
    assert res in [b'key:1', b'key:2', b'key:3']

    # test with encoding param
    res = await redis.randomkey(encoding='utf-8')
    assert res in ['key:1', 'key:2', 'key:3']

    await redis.connection.execute('flushdb')
    res = await redis.randomkey()
    assert res is None


@pytest.mark.run_loop
async def test_rename(redis, server):
    await add(redis, 'foo', 'bar')
    await redis.delete('bar')

    res = await redis.rename('foo', 'bar')
    assert res is True

    with pytest.raises(ReplyError, match='ERR no such key'):
        await redis.rename('foo', 'bar')
    with pytest.raises(TypeError):
        await redis.rename(None, 'bar')
    with pytest.raises(TypeError):
        await redis.rename('foo', None)
    with pytest.raises(ValueError):
        await redis.rename('foo', 'foo')

    if server.version < (3, 2):
        with pytest.raises(ReplyError, match='.* objects are the same'):
            await redis.rename('bar', b'bar')


@pytest.mark.run_loop
async def test_renamenx(redis, server):
    await redis.delete('foo', 'bar')
    await add(redis, 'foo', 123)

    res = await redis.renamenx('foo', 'bar')
    assert res is True

    await add(redis, 'foo', 123)
    res = await redis.renamenx('foo', 'bar')
    assert res is False

    with pytest.raises(ReplyError, match='ERR no such key'):
        await redis.renamenx('baz', 'foo')
    with pytest.raises(TypeError):
        await redis.renamenx(None, 'foo')
    with pytest.raises(TypeError):
        await redis.renamenx('foo', None)
    with pytest.raises(ValueError):
        await redis.renamenx('foo', 'foo')

    if server.version < (3, 2):
        with pytest.raises(ReplyError, match='.* objects are the same'):
            await redis.renamenx('foo', b'foo')


@pytest.mark.run_loop
async def test_restore(redis):
    ok = await redis.set('key', 'value')
    assert ok
    dump = await redis.dump('key')
    assert dump is not None
    ok = await redis.delete('key')
    assert ok
    assert b'OK' == (await redis.restore('key', 0, dump))
    assert (await redis.get('key')) == b'value'


@pytest.redis_version(2, 8, 0, reason='SCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_scan(redis):
    for i in range(1, 11):
        foo_or_bar = 'bar' if i % 3 else 'foo'
        key = 'key:scan:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        await add(redis, key, i)

    cursor, values = await redis.scan()
    # values should be *>=* just in case some other tests left
    # test keys
    assert len(values) >= 10

    cursor, test_values = b'0', []
    while cursor:
        cursor, values = await redis.scan(
            cursor=cursor, match=b'key:scan:foo*')
        test_values.extend(values)
    assert len(test_values) == 3

    cursor, test_values = b'0', []
    while cursor:
        cursor, values = await redis.scan(
            cursor=cursor, match=b'key:scan:bar:*')
        test_values.extend(values)
    assert len(test_values) == 7

    # SCAN family functions do not guarantee that the number of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something
    cursor = b'0'
    test_values = []
    while cursor:
        cursor, values = await redis.scan(cursor=cursor,
                                          match=b'key:scan:*',
                                          count=2)

        test_values.extend(values)
    assert len(test_values) == 10


@pytest.mark.run_loop
async def test_sort(redis):
    async def _make_list(key, items):
        await redis.delete(key)
        for i in items:
            await redis.rpush(key, i)

    await _make_list('a', '4231')
    res = await redis.sort('a')
    assert res == [b'1', b'2', b'3', b'4']

    res = await redis.sort('a', offset=2, count=2)
    assert res == [b'3', b'4']

    res = await redis.sort('a', asc=b'DESC')
    assert res == [b'4', b'3', b'2', b'1']

    await _make_list('a', 'dbca')
    res = await redis.sort(
        'a', asc=b'DESC', alpha=True, offset=2, count=2
    )
    assert res == [b'b', b'a']

    await redis.set('key:1', 10)
    await redis.set('key:2', 4)
    await redis.set('key:3', 7)
    await _make_list('a', '321')

    res = await redis.sort('a', by='key:*')
    assert res == [b'2', b'3', b'1']

    res = await redis.sort('a', by='nosort')
    assert res == [b'3', b'2', b'1']

    res = await redis.sort('a', by='key:*', store='sorted_a')
    assert res == 3
    res = await redis.lrange('sorted_a', 0, -1)
    assert res == [b'2', b'3', b'1']

    await redis.set('value:1', 20)
    await redis.set('value:2', 30)
    await redis.set('value:3', 40)
    res = await redis.sort('a', 'value:*', by='key:*')
    assert res == [b'30', b'40', b'20']

    await redis.hset('data_1', 'weight', 30)
    await redis.hset('data_2', 'weight', 20)
    await redis.hset('data_3', 'weight', 10)
    await redis.hset('hash_1', 'field', 20)
    await redis.hset('hash_2', 'field', 30)
    await redis.hset('hash_3', 'field', 10)
    res = await redis.sort(
        'a', 'hash_*->field', by='data_*->weight'
    )
    assert res == [b'10', b'30', b'20']


@pytest.redis_version(3, 2, 1, reason="TOUCH is available since redis>=3.2.1")
@pytest.mark.run_loop(timeout=20)
async def test_touch(redis, loop):
    await add(redis, 'key', 'val')
    res = 0
    while not res:
        res = await redis.object_idletime('key')
        await asyncio.sleep(.5, loop=loop)
    assert res > 0
    assert await redis.touch('key', 'key', 'key') == 3
    res2 = await redis.object_idletime('key')
    assert 0 <= res2 < res


@pytest.mark.run_loop
async def test_ttl(redis, server):
    await add(redis, 'key', 'val')
    res = await redis.ttl('key')
    assert res == -1
    res = await redis.ttl('non-existent-key')
    if server.version < (2, 8, 0):
        assert res == -1
    else:
        assert res == -2

    await redis.expire('key', 10)
    res = await redis.ttl('key')
    assert res >= 9

    with pytest.raises(TypeError):
        await redis.ttl(None)


@pytest.mark.run_loop
async def test_type(redis):
    await add(redis, 'key', 'val')
    res = await redis.type('key')
    assert res == b'string'

    await redis.delete('key')
    await redis.incr('key')
    res = await redis.type('key')
    assert res == b'string'

    await redis.delete('key')
    await redis.sadd('key', 'val')
    res = await redis.type('key')
    assert res == b'set'

    res = await redis.type('non-existent-key')
    assert res == b'none'

    with pytest.raises(TypeError):
        await redis.type(None)


@pytest.redis_version(2, 8, 0, reason='SCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_iscan(redis):
    full = set()
    foo = set()
    bar = set()
    for i in range(1, 11):
        is_bar = i % 3
        foo_or_bar = 'bar' if is_bar else 'foo'
        key = 'key:scan:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        full.add(key)
        if is_bar:
            bar.add(key)
        else:
            foo.add(key)
        assert await redis.set(key, i) is True

    async def coro(cmd):
        lst = []
        async for i in cmd:
            lst.append(i)
        return lst

    ret = await coro(redis.iscan())
    assert len(ret) >= 10

    ret = await coro(redis.iscan(match='key:scan:*'))
    assert 10 == len(ret)
    assert set(ret) == full

    ret = await coro(redis.iscan(match='key:scan:foo*'))
    assert set(ret) == foo

    ret = await coro(redis.iscan(match='key:scan:bar*'))
    assert set(ret) == bar

    # SCAN family functions do not guarantee that the number of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something

    ret = await coro(redis.iscan(match='key:scan:*', count=2))
    assert 10 == len(ret)
    assert set(ret) == full


@pytest.redis_version(4, 0, 0, reason="UNLINK is available since redis>=4.0.0")
@pytest.mark.run_loop
async def test_unlink(redis):
    await add(redis, 'my-key', 123)
    await add(redis, 'other-key', 123)

    res = await redis.unlink('my-key', 'non-existent-key')
    assert res == 1

    res = await redis.unlink('other-key', 'other-key')
    assert res == 1

    with pytest.raises(TypeError):
        await redis.unlink(None)

    with pytest.raises(TypeError):
        await redis.unlink('my-key', 'my-key', None)


@pytest.redis_version(3, 0, 0, reason="WAIT is available since redis>=3.0.0")
@pytest.mark.run_loop
async def test_wait(redis, loop):
    await add(redis, 'key', 'val1')
    start = await redis.time()
    res = await redis.wait(1, 400)
    end = await redis.time()
    assert res == 0
    assert end - start >= .4

    await add(redis, 'key', 'val2')
    start = await redis.time()
    res = await redis.wait(0, 400)
    end = await redis.time()
    assert res == 0
    assert end - start < .4



from aioredis.errors import ReplyError
from aioredis.errors import MaxClientsError


def test_return_default_class():
    assert isinstance(ReplyError(None), ReplyError)


def test_return_adhoc_class():
    class MyError(ReplyError):
        MATCH_REPLY = "my error"

    assert isinstance(ReplyError("my error"), MyError)


def test_return_max_clients_error():
    assert isinstance(
        ReplyError("ERR max number of clients reached"), MaxClientsError)



import asyncio
import pytest
import aioredis


async def _reader(channel, output, waiter, conn):
    await conn.execute('subscribe', channel)
    ch = conn.pubsub_channels[channel]
    waiter.set_result(conn)
    while await ch.wait_message():
        msg = await ch.get()
        await output.put(msg)


@pytest.mark.run_loop
async def test_publish(create_connection, redis, server, loop):
    out = asyncio.Queue(loop=loop)
    fut = loop.create_future()
    conn = await create_connection(
        server.tcp_address, loop=loop)
    sub = asyncio.ensure_future(_reader('chan:1', out, fut, conn), loop=loop)

    await fut
    await redis.publish('chan:1', 'Hello')
    msg = await out.get()
    assert msg == b'Hello'

    sub.cancel()


@pytest.mark.run_loop
async def test_publish_json(create_connection, redis, server, loop):
    out = asyncio.Queue(loop=loop)
    fut = loop.create_future()
    conn = await create_connection(
        server.tcp_address, loop=loop)
    sub = asyncio.ensure_future(_reader('chan:1', out, fut, conn), loop=loop)

    await fut

    res = await redis.publish_json('chan:1', {"Hello": "world"})
    assert res == 1    # recievers

    msg = await out.get()
    assert msg == b'{"Hello": "world"}'
    sub.cancel()


@pytest.mark.run_loop
async def test_subscribe(redis):
    res = await redis.subscribe('chan:1', 'chan:2')
    assert redis.in_pubsub == 2

    ch1 = redis.channels['chan:1']
    ch2 = redis.channels['chan:2']

    assert res == [ch1, ch2]
    assert ch1.is_pattern is False
    assert ch2.is_pattern is False

    res = await redis.unsubscribe('chan:1', 'chan:2')
    assert res == [[b'unsubscribe', b'chan:1', 1],
                   [b'unsubscribe', b'chan:2', 0]]


@pytest.mark.parametrize('create_redis', [
    pytest.param(aioredis.create_redis_pool, id='pool'),
])
@pytest.mark.run_loop
async def test_subscribe_empty_pool(create_redis, server, loop, _closable):
    redis = await create_redis(server.tcp_address, loop=loop)
    _closable(redis)
    await redis.connection.clear()

    res = await redis.subscribe('chan:1', 'chan:2')
    assert redis.in_pubsub == 2

    ch1 = redis.channels['chan:1']
    ch2 = redis.channels['chan:2']

    assert res == [ch1, ch2]
    assert ch1.is_pattern is False
    assert ch2.is_pattern is False

    res = await redis.unsubscribe('chan:1', 'chan:2')
    assert res == [[b'unsubscribe', b'chan:1', 1],
                   [b'unsubscribe', b'chan:2', 0]]


@pytest.mark.run_loop
async def test_psubscribe(redis, create_redis, server, loop):
    sub = redis
    res = await sub.psubscribe('patt:*', 'chan:*')
    assert sub.in_pubsub == 2

    pat1 = sub.patterns['patt:*']
    pat2 = sub.patterns['chan:*']
    assert res == [pat1, pat2]

    pub = await create_redis(
        server.tcp_address, loop=loop)
    await pub.publish_json('chan:123', {"Hello": "World"})
    res = await pat2.get_json()
    assert res == (b'chan:123', {"Hello": "World"})

    res = await sub.punsubscribe('patt:*', 'patt:*', 'chan:*')
    assert res == [[b'punsubscribe', b'patt:*', 1],
                   [b'punsubscribe', b'patt:*', 1],
                   [b'punsubscribe', b'chan:*', 0],
                   ]


@pytest.mark.parametrize('create_redis', [
    pytest.param(aioredis.create_redis_pool, id='pool'),
])
@pytest.mark.run_loop
async def test_psubscribe_empty_pool(create_redis, server, loop, _closable):
    sub = await create_redis(server.tcp_address, loop=loop)
    pub = await create_redis(server.tcp_address, loop=loop)
    _closable(sub)
    _closable(pub)
    await sub.connection.clear()
    res = await sub.psubscribe('patt:*', 'chan:*')
    assert sub.in_pubsub == 2

    pat1 = sub.patterns['patt:*']
    pat2 = sub.patterns['chan:*']
    assert res == [pat1, pat2]

    await pub.publish_json('chan:123', {"Hello": "World"})
    res = await pat2.get_json()
    assert res == (b'chan:123', {"Hello": "World"})

    res = await sub.punsubscribe('patt:*', 'patt:*', 'chan:*')
    assert res == [[b'punsubscribe', b'patt:*', 1],
                   [b'punsubscribe', b'patt:*', 1],
                   [b'punsubscribe', b'chan:*', 0],
                   ]


@pytest.redis_version(
    2, 8, 0, reason='PUBSUB CHANNELS is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_pubsub_channels(create_redis, server, loop):
    redis = await create_redis(
        server.tcp_address, loop=loop)
    res = await redis.pubsub_channels()
    assert res == []

    res = await redis.pubsub_channels('chan:*')
    assert res == []

    sub = await create_redis(
        server.tcp_address, loop=loop)
    await sub.subscribe('chan:1')

    res = await redis.pubsub_channels()
    assert res == [b'chan:1']

    res = await redis.pubsub_channels('ch*')
    assert res == [b'chan:1']

    await sub.unsubscribe('chan:1')
    await sub.psubscribe('chan:*')

    res = await redis.pubsub_channels()
    assert res == []


@pytest.redis_version(
    2, 8, 0, reason='PUBSUB NUMSUB is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_pubsub_numsub(create_redis, server, loop):
    redis = await create_redis(
        server.tcp_address, loop=loop)
    res = await redis.pubsub_numsub()
    assert res == {}

    res = await redis.pubsub_numsub('chan:1')
    assert res == {b'chan:1': 0}

    sub = await create_redis(
        server.tcp_address, loop=loop)
    await sub.subscribe('chan:1')

    res = await redis.pubsub_numsub()
    assert res == {}

    res = await redis.pubsub_numsub('chan:1')
    assert res == {b'chan:1': 1}

    res = await redis.pubsub_numsub('chan:2')
    assert res == {b'chan:2': 0}

    res = await redis.pubsub_numsub('chan:1', 'chan:2')
    assert res == {b'chan:1': 1, b'chan:2': 0}

    await sub.unsubscribe('chan:1')
    await sub.psubscribe('chan:*')

    res = await redis.pubsub_numsub()
    assert res == {}


@pytest.redis_version(
    2, 8, 0, reason='PUBSUB NUMPAT is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_pubsub_numpat(create_redis, server, loop, redis):
    sub = await create_redis(
        server.tcp_address, loop=loop)

    res = await redis.pubsub_numpat()
    assert res == 0

    await sub.subscribe('chan:1')
    res = await redis.pubsub_numpat()
    assert res == 0

    await sub.psubscribe('chan:*')
    res = await redis.pubsub_numpat()
    assert res == 1


@pytest.mark.run_loop
async def test_close_pubsub_channels(redis, loop):
    ch, = await redis.subscribe('chan:1')

    async def waiter(ch):
        assert not await ch.wait_message()

    tsk = asyncio.ensure_future(waiter(ch), loop=loop)
    redis.close()
    await redis.wait_closed()
    await tsk


@pytest.mark.run_loop
async def test_close_pubsub_patterns(redis, loop):
    ch, = await redis.psubscribe('chan:*')

    async def waiter(ch):
        assert not await ch.wait_message()

    tsk = asyncio.ensure_future(waiter(ch), loop=loop)
    redis.close()
    await redis.wait_closed()
    await tsk


@pytest.mark.run_loop
async def test_close_cancelled_pubsub_channel(redis, loop):
    ch, = await redis.subscribe('chan:1')

    async def waiter(ch):
        with pytest.raises(asyncio.CancelledError):
            await ch.wait_message()

    tsk = asyncio.ensure_future(waiter(ch), loop=loop)
    await asyncio.sleep(0, loop=loop)
    tsk.cancel()


@pytest.mark.run_loop
async def test_channel_get_after_close(create_redis, loop, server):
    sub = await create_redis(
        server.tcp_address, loop=loop)
    pub = await create_redis(
        server.tcp_address, loop=loop)
    ch, = await sub.subscribe('chan:1')

    await pub.publish('chan:1', 'message')
    assert await ch.get() == b'message'
    loop.call_soon(sub.close)
    assert await ch.get() is None
    with pytest.raises(aioredis.ChannelClosedError):
        assert await ch.get()


@pytest.mark.run_loop
async def test_subscribe_concurrency(create_redis, server, loop):
    sub = await create_redis(
        server.tcp_address, loop=loop)
    pub = await create_redis(
        server.tcp_address, loop=loop)

    async def subscribe(*args):
        return await sub.subscribe(*args)

    async def publish(*args):
        await asyncio.sleep(0, loop=loop)
        return await pub.publish(*args)

    res = await asyncio.gather(
        subscribe('channel:0'),
        publish('channel:0', 'Hello'),
        subscribe('channel:1'),
        loop=loop)
    (ch1,), subs, (ch2,) = res

    assert ch1.name == b'channel:0'
    assert subs == 1
    assert ch2.name == b'channel:1'


@pytest.redis_version(
    3, 2, 0, reason='PUBSUB PING is available since redis>=3.2.0')
@pytest.mark.run_loop
async def test_pubsub_ping(redis):
    await redis.subscribe('chan:1', 'chan:2')

    res = await redis.ping()
    assert res == b'PONG'
    res = await redis.ping('Hello')
    assert res == b'Hello'
    res = await redis.ping('Hello', encoding='utf-8')
    assert res == 'Hello'

    await redis.unsubscribe('chan:1', 'chan:2')


@pytest.mark.run_loop
async def test_pubsub_channel_iter(create_redis, server, loop):
    sub = await create_redis(server.tcp_address, loop=loop)
    pub = await create_redis(server.tcp_address, loop=loop)

    ch, = await sub.subscribe('chan:1')

    async def coro(ch):
        lst = []
        async for msg in ch.iter():
            lst.append(msg)
        return lst

    tsk = asyncio.ensure_future(coro(ch), loop=loop)
    await pub.publish_json('chan:1', {'Hello': 'World'})
    await pub.publish_json('chan:1', ['message'])
    await asyncio.sleep(0, loop=loop)
    ch.close()
    assert await tsk == [b'{"Hello": "World"}', b'["message"]']



from aioredis.util import wait_ok


class ScriptingCommandsMixin:
    """Set commands mixin.

    For commands details see: http://redis.io/commands#scripting
    """

    def eval(self, script, keys=[], args=[]):
        """Execute a Lua script server side."""
        return self.execute(b'EVAL', script, len(keys), *(keys + args))

    def evalsha(self, digest, keys=[], args=[]):
        """Execute a Lua script server side by its SHA1 digest."""
        return self.execute(b'EVALSHA', digest, len(keys), *(keys + args))

    def script_exists(self, digest, *digests):
        """Check existence of scripts in the script cache."""
        return self.execute(b'SCRIPT', b'EXISTS', digest, *digests)

    def script_kill(self):
        """Kill the script currently in execution."""
        fut = self.execute(b'SCRIPT', b'KILL')
        return wait_ok(fut)

    def script_flush(self):
        """Remove all the scripts from the script cache."""
        fut = self.execute(b"SCRIPT",  b"FLUSH")
        return wait_ok(fut)

    def script_load(self, script):
        """Load the specified Lua script into the script cache."""
        return self.execute(b"SCRIPT",  b"LOAD", script)



import time
import pytest
import sys

from unittest import mock

from aioredis import ReplyError


@pytest.mark.run_loop
async def test_client_list(redis, server, request):
    name = request.node.callspec.id
    assert (await redis.client_setname(name))
    res = await redis.client_list()
    assert isinstance(res, list)
    res = [dict(i._asdict()) for i in res]
    expected = {
        'addr': mock.ANY,
        'fd': mock.ANY,
        'age': mock.ANY,
        'idle': mock.ANY,
        'flags': 'N',
        'db': '0',
        'sub': '0',
        'psub': '0',
        'multi': '-1',
        'qbuf': '0',
        'qbuf_free': mock.ANY,
        'obl': '0',
        'oll': '0',
        'omem': '0',
        'events': 'r',
        'cmd': 'client',
        'name': name,
        }
    if server.version >= (2, 8, 12):
        expected['id'] = mock.ANY
    if server.version >= (5, ):
        expected['qbuf'] = '26'
    assert expected in res


@pytest.mark.run_loop
@pytest.mark.skipif(sys.platform == 'win32',
                    reason="No unixsocket on Windows")
async def test_client_list__unixsocket(create_redis, loop, server, request):
    redis = await create_redis(server.unixsocket, loop=loop)
    name = request.node.callspec.id
    assert (await redis.client_setname(name))
    res = await redis.client_list()
    info = [dict(i._asdict()) for i in res]
    expected = {
        'addr': '{}:0'.format(server.unixsocket),
        'fd': mock.ANY,
        'age': mock.ANY,
        'idle': mock.ANY,
        'flags': 'U',   # Conneted via unix socket
        'db': '0',
        'sub': '0',
        'psub': '0',
        'multi': '-1',
        'qbuf': '0',
        'qbuf_free': mock.ANY,
        'obl': '0',
        'oll': '0',
        'omem': '0',
        'events': 'r',
        'cmd': 'client',
        'name': name,
        }
    if server.version >= (2, 8, 12):
        expected['id'] = mock.ANY
    if server.version >= (5, ):
        expected['qbuf'] = '26'
    assert expected in info


@pytest.mark.run_loop
@pytest.redis_version(
    2, 9, 50, reason='CLIENT PAUSE is available since redis >= 2.9.50')
async def test_client_pause(redis):
    ts = time.time()
    res = await redis.client_pause(2000)
    assert res is True
    await redis.ping()
    assert int(time.time() - ts) >= 2

    with pytest.raises(TypeError):
        await redis.client_pause(2.0)
    with pytest.raises(ValueError):
        await redis.client_pause(-1)


@pytest.mark.run_loop
async def test_client_getname(redis):
    res = await redis.client_getname()
    assert res is None
    ok = await redis.client_setname('TestClient')
    assert ok is True

    res = await redis.client_getname()
    assert res == b'TestClient'
    res = await redis.client_getname(encoding='utf-8')
    assert res == 'TestClient'


@pytest.redis_version(2, 8, 13, reason="available since Redis 2.8.13")
@pytest.mark.run_loop
async def test_command(redis):
    res = await redis.command()
    assert isinstance(res, list)
    assert len(res) > 0


@pytest.redis_version(2, 8, 13, reason="available since Redis 2.8.13")
@pytest.mark.run_loop
async def test_command_count(redis):
    res = await redis.command_count()
    assert res > 0


@pytest.redis_version(3, 0, 0, reason="available since Redis 3.0.0")
@pytest.mark.run_loop
async def test_command_getkeys(redis):
    res = await redis.command_getkeys('get', 'key')
    assert res == ['key']
    res = await redis.command_getkeys('get', 'key', encoding=None)
    assert res == [b'key']
    res = await redis.command_getkeys('mset', 'k1', 'v1', 'k2', 'v2')
    assert res == ['k1', 'k2']
    res = await redis.command_getkeys('mset', 'k1', 'v1', 'k2')
    assert res == ['k1', 'k2']

    with pytest.raises(ReplyError):
        assert (await redis.command_getkeys('get'))
    with pytest.raises(TypeError):
        assert not (await redis.command_getkeys(None))


@pytest.redis_version(2, 8, 13, reason="available since Redis 2.8.13")
@pytest.mark.run_loop
async def test_command_info(redis):
    res = await redis.command_info('get')
    assert res == [
        ['get', 2, ['readonly', 'fast'], 1, 1, 1],
    ]

    res = await redis.command_info("unknown-command")
    assert res == [None]
    res = await redis.command_info("unknown-command", "unknown-commnad")
    assert res == [None, None]


@pytest.mark.run_loop
async def test_config_get(redis, server):
    res = await redis.config_get('port')
    assert res == {'port': str(server.tcp_address.port)}

    res = await redis.config_get()
    assert len(res) > 0

    res = await redis.config_get('unknown_parameter')
    assert res == {}

    with pytest.raises(TypeError):
        await redis.config_get(b'port')


@pytest.mark.run_loop
async def test_config_rewrite(redis):
    with pytest.raises(ReplyError):
        await redis.config_rewrite()


@pytest.mark.run_loop
async def test_config_set(redis):
    cur_value = await redis.config_get('slave-read-only')
    res = await redis.config_set('slave-read-only', 'no')
    assert res is True
    res = await redis.config_set(
        'slave-read-only', cur_value['slave-read-only'])
    assert res is True

    with pytest.raises(ReplyError, match="Unsupported CONFIG parameter"):
        await redis.config_set('databases', 100)
    with pytest.raises(TypeError):
        await redis.config_set(100, 'databases')


# @pytest.mark.run_loop
# @pytest.mark.skip("Not implemented")
# def test_config_resetstat():
#     pass

@pytest.mark.run_loop
async def test_debug_object(redis):
    with pytest.raises(ReplyError):
        assert (await redis.debug_object('key')) is None

    ok = await redis.set('key', 'value')
    assert ok
    res = await redis.debug_object('key')
    assert res is not None


@pytest.mark.run_loop
async def test_debug_sleep(redis):
    t1 = await redis.time()
    ok = await redis.debug_sleep(2)
    assert ok
    t2 = await redis.time()
    assert t2 - t1 >= 2


@pytest.mark.run_loop
async def test_dbsize(redis):
    res = await redis.dbsize()
    assert res == 0

    await redis.set('key', 'value')

    res = await redis.dbsize()
    assert res > 0

    await redis.flushdb()
    res = await redis.dbsize()
    assert res == 0
    await redis.set('key', 'value')
    res = await redis.dbsize()
    assert res == 1


@pytest.mark.run_loop
async def test_info(redis):
    res = await redis.info()
    assert isinstance(res, dict)

    res = await redis.info('all')
    assert isinstance(res, dict)

    with pytest.raises(ValueError):
        await redis.info('')


@pytest.mark.run_loop
async def test_lastsave(redis):
    res = await redis.lastsave()
    assert res > 0


@pytest.mark.run_loop
@pytest.redis_version(2, 8, 12, reason='ROLE is available since redis>=2.8.12')
async def test_role(redis):
    res = await redis.role()
    assert dict(res._asdict()) == {
        'role': 'master',
        'replication_offset': mock.ANY,
        'slaves': [],
        }


@pytest.mark.run_loop
async def test_save(redis):
    res = await redis.dbsize()
    assert res == 0
    t1 = await redis.lastsave()
    ok = await redis.save()
    assert ok
    t2 = await redis.lastsave()
    assert t2 >= t1


@pytest.mark.run_loop
async def test_time(redis):
    res = await redis.time()
    assert isinstance(res, float)
    pytest.assert_almost_equal(int(res), int(time.time()), delta=10)


@pytest.mark.run_loop
async def test_time_with_encoding(create_redis, server, loop):
    redis = await create_redis(server.tcp_address, loop=loop,
                               encoding='utf-8')
    res = await redis.time()
    assert isinstance(res, float)
    pytest.assert_almost_equal(int(res), int(time.time()), delta=10)


@pytest.mark.run_loop
async def test_slowlog_len(redis):
    res = await redis.slowlog_len()
    assert res >= 0


@pytest.mark.run_loop
async def test_slowlog_get(redis):
    res = await redis.slowlog_get()
    assert isinstance(res, list)
    assert len(res) >= 0

    res = await redis.slowlog_get(2)
    assert isinstance(res, list)
    assert 0 <= len(res) <= 2

    with pytest.raises(TypeError):
        assert not (await redis.slowlog_get(1.2))
    with pytest.raises(TypeError):
        assert not (await redis.slowlog_get('1'))


@pytest.mark.run_loop
async def test_slowlog_reset(redis):
    ok = await redis.slowlog_reset()
    assert ok is True



from aioredis.util import _NOTSET, wait_ok


class ListCommandsMixin:
    """List commands mixin.

    For commands details see: http://redis.io/commands#list
    """

    def blpop(self, key, *keys, timeout=0, encoding=_NOTSET):
        """Remove and get the first element in a list, or block until
        one is available.

        :raises TypeError: if timeout is not int
        :raises ValueError: if timeout is less than 0
        """
        if not isinstance(timeout, int):
            raise TypeError("timeout argument must be int")
        if timeout < 0:
            raise ValueError("timeout must be greater equal 0")
        args = keys + (timeout,)
        return self.execute(b'BLPOP', key, *args, encoding=encoding)

    def brpop(self, key, *keys, timeout=0, encoding=_NOTSET):
        """Remove and get the last element in a list, or block until one
        is available.

        :raises TypeError: if timeout is not int
        :raises ValueError: if timeout is less than 0
        """
        if not isinstance(timeout, int):
            raise TypeError("timeout argument must be int")
        if timeout < 0:
            raise ValueError("timeout must be greater equal 0")
        args = keys + (timeout,)
        return self.execute(b'BRPOP', key, *args, encoding=encoding)

    def brpoplpush(self, sourcekey, destkey, timeout=0, encoding=_NOTSET):
        """Remove and get the last element in a list, or block until one
        is available.

        :raises TypeError: if timeout is not int
        :raises ValueError: if timeout is less than 0
        """
        if not isinstance(timeout, int):
            raise TypeError("timeout argument must be int")
        if timeout < 0:
            raise ValueError("timeout must be greater equal 0")
        return self.execute(b'BRPOPLPUSH', sourcekey, destkey, timeout,
                            encoding=encoding)

    def lindex(self, key, index, *, encoding=_NOTSET):
        """Get an element from a list by its index.

        :raises TypeError: if index is not int
        """
        if not isinstance(index, int):
            raise TypeError("index argument must be int")
        return self.execute(b'LINDEX', key, index, encoding=encoding)

    def linsert(self, key, pivot, value, before=False):
        """Inserts value in the list stored at key either before or
        after the reference value pivot.
        """
        where = b'AFTER' if not before else b'BEFORE'
        return self.execute(b'LINSERT', key, where, pivot, value)

    def llen(self, key):
        """Returns the length of the list stored at key."""
        return self.execute(b'LLEN', key)

    def lpop(self, key, *, encoding=_NOTSET):
        """Removes and returns the first element of the list stored at key."""
        return self.execute(b'LPOP', key, encoding=encoding)

    def lpush(self, key, value, *values):
        """Insert all the specified values at the head of the list
        stored at key.
        """
        return self.execute(b'LPUSH', key, value, *values)

    def lpushx(self, key, value):
        """Inserts value at the head of the list stored at key, only if key
        already exists and holds a list.
        """
        return self.execute(b'LPUSHX', key, value)

    def lrange(self, key, start, stop, *, encoding=_NOTSET):
        """Returns the specified elements of the list stored at key.

        :raises TypeError: if start or stop is not int
        """
        if not isinstance(start, int):
            raise TypeError("start argument must be int")
        if not isinstance(stop, int):
            raise TypeError("stop argument must be int")
        return self.execute(b'LRANGE', key, start, stop, encoding=encoding)

    def lrem(self, key, count, value):
        """Removes the first count occurrences of elements equal to value
        from the list stored at key.

        :raises TypeError: if count is not int
        """
        if not isinstance(count, int):
            raise TypeError("count argument must be int")
        return self.execute(b'LREM', key, count, value)

    def lset(self, key, index, value):
        """Sets the list element at index to value.

        :raises TypeError: if index is not int
        """
        if not isinstance(index, int):
            raise TypeError("index argument must be int")
        return self.execute(b'LSET', key, index, value)

    def ltrim(self, key, start, stop):
        """Trim an existing list so that it will contain only the specified
        range of elements specified.

        :raises TypeError: if start or stop is not int
        """
        if not isinstance(start, int):
            raise TypeError("start argument must be int")
        if not isinstance(stop, int):
            raise TypeError("stop argument must be int")
        fut = self.execute(b'LTRIM', key, start, stop)
        return wait_ok(fut)

    def rpop(self, key, *, encoding=_NOTSET):
        """Removes and returns the last element of the list stored at key."""
        return self.execute(b'RPOP', key, encoding=encoding)

    def rpoplpush(self, sourcekey, destkey, *, encoding=_NOTSET):
        """Atomically returns and removes the last element (tail) of the
        list stored at source, and pushes the element at the first element
        (head) of the list stored at destination.
        """
        return self.execute(b'RPOPLPUSH', sourcekey, destkey,
                            encoding=encoding)

    def rpush(self, key, value, *values):
        """Insert all the specified values at the tail of the list
        stored at key.
        """
        return self.execute(b'RPUSH', key, value, *values)

    def rpushx(self, key, value):
        """Inserts value at the tail of the list stored at key, only if
        key already exists and holds a list.
        """
        return self.execute(b'RPUSHX', key, value)



import asyncio
import aioredis


async def main():
    conn = await aioredis.create_connection(
        'redis://localhost', encoding='utf-8')

    ok = await conn.execute('set', 'my-key', 'some value')
    assert ok == 'OK', ok

    str_value = await conn.execute('get', 'my-key')
    raw_value = await conn.execute('get', 'my-key', encoding=None)
    assert str_value == 'some value'
    assert raw_value == b'some value'

    print('str value:', str_value)
    print('raw value:', raw_value)

    # optionally close connection
    conn.close()
    await conn.wait_closed()


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())



import json

from aioredis.util import wait_make_dict


class PubSubCommandsMixin:
    """Pub/Sub commands mixin.

    For commands details see: http://redis.io/commands/#pubsub
    """

    def publish(self, channel, message):
        """Post a message to channel."""
        return self.execute(b'PUBLISH', channel, message)

    def publish_json(self, channel, obj):
        """Post a JSON-encoded message to channel."""
        return self.publish(channel, json.dumps(obj))

    def subscribe(self, channel, *channels):
        """Switch connection to Pub/Sub mode and
        subscribe to specified channels.

        Arguments can be instances of :class:`~aioredis.Channel`.

        Returns :func:`asyncio.gather()` coroutine which when done will return
        a list of :class:`~aioredis.Channel` objects.
        """
        conn = self._pool_or_conn
        return wait_return_channels(
            conn.execute_pubsub(b'SUBSCRIBE', channel, *channels),
            conn, 'pubsub_channels')

    def unsubscribe(self, channel, *channels):
        """Unsubscribe from specific channels.

        Arguments can be instances of :class:`~aioredis.Channel`.
        """
        conn = self._pool_or_conn
        return conn.execute_pubsub(b'UNSUBSCRIBE', channel, *channels)

    def psubscribe(self, pattern, *patterns):
        """Switch connection to Pub/Sub mode and
        subscribe to specified patterns.

        Arguments can be instances of :class:`~aioredis.Channel`.

        Returns :func:`asyncio.gather()` coroutine which when done will return
        a list of subscribed :class:`~aioredis.Channel` objects with
        ``is_pattern`` property set to ``True``.
        """
        conn = self._pool_or_conn
        return wait_return_channels(
            conn.execute_pubsub(b'PSUBSCRIBE', pattern, *patterns),
            conn, 'pubsub_patterns')

    def punsubscribe(self, pattern, *patterns):
        """Unsubscribe from specific patterns.

        Arguments can be instances of :class:`~aioredis.Channel`.
        """
        conn = self._pool_or_conn
        return conn.execute_pubsub(b'PUNSUBSCRIBE', pattern, *patterns)

    def pubsub_channels(self, pattern=None):
        """Lists the currently active channels."""
        args = [b'PUBSUB', b'CHANNELS']
        if pattern is not None:
            args.append(pattern)
        return self.execute(*args)

    def pubsub_numsub(self, *channels):
        """Returns the number of subscribers for the specified channels."""
        return wait_make_dict(self.execute(
            b'PUBSUB', b'NUMSUB', *channels))

    def pubsub_numpat(self):
        """Returns the number of subscriptions to patterns."""
        return self.execute(b'PUBSUB', b'NUMPAT')

    @property
    def channels(self):
        """Returns read-only channels dict.

        See :attr:`~aioredis.RedisConnection.pubsub_channels`
        """
        return self._pool_or_conn.pubsub_channels

    @property
    def patterns(self):
        """Returns read-only patterns dict.

        See :attr:`~aioredis.RedisConnection.pubsub_patterns`
        """
        return self._pool_or_conn.pubsub_patterns

    @property
    def in_pubsub(self):
        """Indicates that connection is in PUB/SUB mode.

        Provides the number of subscribed channels.
        """
        return self._pool_or_conn.in_pubsub


async def wait_return_channels(fut, conn, field):
    res = await fut
    channels_dict = getattr(conn, field)
    return [channels_dict[name] for cmd, name, count in res]



import pytest
import asyncio
import json
import sys

from unittest import mock

from aioredis import ChannelClosedError
from aioredis.abc import AbcChannel
from aioredis.pubsub import Receiver, _Sender


def test_listener_channel(loop):
    mpsc = Receiver(loop=loop)
    assert not mpsc.is_active

    ch_a = mpsc.channel("channel:1")
    assert isinstance(ch_a, AbcChannel)
    assert mpsc.is_active

    ch_b = mpsc.channel('channel:1')
    assert ch_a is ch_b
    assert ch_a.name == ch_b.name
    assert ch_a.is_pattern == ch_b.is_pattern
    assert mpsc.is_active

    # remember id; drop refs to objects and create new one;
    ch_a.close()
    assert not ch_a.is_active

    assert not mpsc.is_active
    ch = mpsc.channel("channel:1")
    assert ch is not ch_a

    assert dict(mpsc.channels) == {b'channel:1': ch}
    assert dict(mpsc.patterns) == {}


def test_listener_pattern(loop):
    mpsc = Receiver(loop=loop)
    assert not mpsc.is_active

    ch_a = mpsc.pattern("*")
    assert isinstance(ch_a, AbcChannel)
    assert mpsc.is_active

    ch_b = mpsc.pattern('*')
    assert ch_a is ch_b
    assert ch_a.name == ch_b.name
    assert ch_a.is_pattern == ch_b.is_pattern
    assert mpsc.is_active

    # remember id; drop refs to objects and create new one;
    ch_a.close()
    assert not ch_a.is_active

    assert not mpsc.is_active
    ch = mpsc.pattern("*")
    assert ch is not ch_a

    assert dict(mpsc.channels) == {}
    assert dict(mpsc.patterns) == {b'*': ch}


@pytest.mark.run_loop
async def test_sender(loop):
    receiver = mock.Mock()

    sender = _Sender(receiver, 'name', is_pattern=False)
    assert isinstance(sender, AbcChannel)
    assert sender.name == b'name'
    assert sender.is_pattern is False
    assert sender.is_active is True

    with pytest.raises(RuntimeError):
        await sender.get()
    assert receiver.mock_calls == []

    sender.put_nowait(b'some data')
    assert receiver.mock_calls == [
        mock.call._put_nowait(b'some data', sender=sender),
        ]


def test_sender_close():
    receiver = mock.Mock()
    sender = _Sender(receiver, 'name', is_pattern=False)
    sender.close()
    assert receiver.mock_calls == [mock.call._close(sender, exc=None)]
    sender.close()
    assert receiver.mock_calls == [mock.call._close(sender, exc=None)]
    receiver.reset_mock()
    assert receiver.mock_calls == []
    sender.close()
    assert receiver.mock_calls == []


@pytest.mark.run_loop
async def test_subscriptions(create_connection, server, loop):
    sub = await create_connection(server.tcp_address, loop=loop)
    pub = await create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    await sub.execute_pubsub('subscribe',
                             mpsc.channel('channel:1'),
                             mpsc.channel('channel:3'))
    res = await pub.execute("publish", "channel:3", "Hello world")
    assert res == 1
    res = await pub.execute("publish", "channel:1", "Hello world")
    assert res == 1
    assert mpsc.is_active

    ch, msg = await mpsc.get()
    assert ch.name == b'channel:3'
    assert not ch.is_pattern
    assert msg == b"Hello world"

    ch, msg = await mpsc.get()
    assert ch.name == b'channel:1'
    assert not ch.is_pattern
    assert msg == b"Hello world"


@pytest.mark.run_loop
async def test_unsubscribe(create_connection, server, loop):
    sub = await create_connection(server.tcp_address, loop=loop)
    pub = await create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    await sub.execute_pubsub('subscribe',
                             mpsc.channel('channel:1'),
                             mpsc.channel('channel:3'))
    res = await pub.execute("publish", "channel:3", "Hello world")
    assert res == 1
    res = await pub.execute("publish", "channel:1", "Hello world")
    assert res == 1
    assert mpsc.is_active

    assert (await mpsc.wait_message()) is True
    ch, msg = await mpsc.get()
    assert ch.name == b'channel:3'
    assert not ch.is_pattern
    assert msg == b"Hello world"

    assert (await mpsc.wait_message()) is True
    ch, msg = await mpsc.get()
    assert ch.name == b'channel:1'
    assert not ch.is_pattern
    assert msg == b"Hello world"

    await sub.execute_pubsub('unsubscribe', 'channel:1')
    assert mpsc.is_active

    res = await pub.execute("publish", "channel:3", "message")
    assert res == 1
    assert (await mpsc.wait_message()) is True
    ch, msg = await mpsc.get()
    assert ch.name == b'channel:3'
    assert not ch.is_pattern
    assert msg == b"message"

    waiter = asyncio.ensure_future(mpsc.get(), loop=loop)
    await sub.execute_pubsub('unsubscribe', 'channel:3')
    assert not mpsc.is_active
    assert await waiter is None


@pytest.mark.run_loop
async def test_stopped(create_connection, server, loop):
    sub = await create_connection(server.tcp_address, loop=loop)
    pub = await create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    await sub.execute_pubsub('subscribe', mpsc.channel('channel:1'))
    assert mpsc.is_active
    mpsc.stop()

    with pytest.logs('aioredis', 'DEBUG') as cm:
        await pub.execute('publish', 'channel:1', b'Hello')
        await asyncio.sleep(0, loop=loop)

    assert len(cm.output) == 1
    # Receiver must have 1 EndOfStream message
    warn_messaege = (
        "WARNING:aioredis:Pub/Sub listener message after stop: "
        "sender: <_Sender name:b'channel:1', is_pattern:False, receiver:"
        "<Receiver is_active:False, senders:1, qsize:0>>, data: b'Hello'"
    )
    assert cm.output == [warn_messaege]

    # assert (await mpsc.get()) is None
    with pytest.raises(ChannelClosedError):
        await mpsc.get()
    res = await mpsc.wait_message()
    assert res is False


@pytest.mark.run_loop
async def test_wait_message(create_connection, server, loop):
    sub = await create_connection(server.tcp_address, loop=loop)
    pub = await create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    await sub.execute_pubsub('subscribe', mpsc.channel('channel:1'))
    fut = asyncio.ensure_future(mpsc.wait_message(), loop=loop)
    assert not fut.done()
    await asyncio.sleep(0, loop=loop)
    assert not fut.done()

    await pub.execute('publish', 'channel:1', 'hello')
    await asyncio.sleep(0, loop=loop)  # read in connection
    await asyncio.sleep(0, loop=loop)  # call Future.set_result
    assert fut.done()
    res = await fut
    assert res is True


@pytest.mark.run_loop
async def test_decode_message(loop):
    mpsc = Receiver(loop)
    ch = mpsc.channel('channel:1')
    ch.put_nowait(b'Some data')

    res = await mpsc.get(encoding='utf-8')
    assert isinstance(res[0], _Sender)
    assert res[1] == 'Some data'

    ch.put_nowait('{"hello": "world"}')
    res = await mpsc.get(decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == {'hello': 'world'}

    ch.put_nowait(b'{"hello": "world"}')
    res = await mpsc.get(encoding='utf-8', decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == {'hello': 'world'}


@pytest.mark.skipif(sys.version_info >= (3, 6),
                    reason="json.loads accept bytes since Python 3.6")
@pytest.mark.run_loop
async def test_decode_message_error(loop):
    mpsc = Receiver(loop)
    ch = mpsc.channel('channel:1')

    ch.put_nowait(b'{"hello": "world"}')
    unexpected = (mock.ANY, {'hello': 'world'})
    with pytest.raises(TypeError):
        assert (await mpsc.get(decoder=json.loads)) == unexpected

    ch = mpsc.pattern('*')
    ch.put_nowait((b'channel', b'{"hello": "world"}'))
    unexpected = (mock.ANY, b'channel', {'hello': 'world'})
    with pytest.raises(TypeError):
        assert (await mpsc.get(decoder=json.loads)) == unexpected


@pytest.mark.run_loop
async def test_decode_message_for_pattern(loop):
    mpsc = Receiver(loop)
    ch = mpsc.pattern('*')
    ch.put_nowait((b'channel', b'Some data'))

    res = await mpsc.get(encoding='utf-8')
    assert isinstance(res[0], _Sender)
    assert res[1] == (b'channel', 'Some data')

    ch.put_nowait((b'channel', '{"hello": "world"}'))
    res = await mpsc.get(decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == (b'channel', {'hello': 'world'})

    ch.put_nowait((b'channel', b'{"hello": "world"}'))
    res = await mpsc.get(encoding='utf-8', decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == (b'channel', {'hello': 'world'})


@pytest.mark.run_loop
async def test_pubsub_receiver_iter(create_redis, server, loop):
    sub = await create_redis(server.tcp_address, loop=loop)
    pub = await create_redis(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)

    async def coro(mpsc):
        lst = []
        async for msg in mpsc.iter():
            lst.append(msg)
        return lst

    tsk = asyncio.ensure_future(coro(mpsc), loop=loop)
    snd1, = await sub.subscribe(mpsc.channel('chan:1'))
    snd2, = await sub.subscribe(mpsc.channel('chan:2'))
    snd3, = await sub.psubscribe(mpsc.pattern('chan:*'))

    subscribers = await pub.publish_json('chan:1', {'Hello': 'World'})
    assert subscribers > 1
    subscribers = await pub.publish_json('chan:2', ['message'])
    assert subscribers > 1
    loop.call_later(0, mpsc.stop)
    # await asyncio.sleep(0, loop=loop)
    assert await tsk == [
        (snd1, b'{"Hello": "World"}'),
        (snd3, (b'chan:1', b'{"Hello": "World"}')),
        (snd2, b'["message"]'),
        (snd3, (b'chan:2', b'["message"]')),
        ]
    assert not mpsc.is_active


@pytest.mark.run_loop(timeout=5)
async def test_pubsub_receiver_call_stop_with_empty_queue(
        create_redis, server, loop):
    sub = await create_redis(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)

    # FIXME: currently at least one subscriber is needed
    snd1, = await sub.subscribe(mpsc.channel('chan:1'))

    now = loop.time()
    loop.call_later(.5, mpsc.stop)
    async for i in mpsc.iter():  # noqa (flake8 bug with async for)
        assert False, "StopAsyncIteration not raised"
    dt = loop.time() - now
    assert dt <= 1.5
    assert not mpsc.is_active


@pytest.mark.run_loop
async def test_pubsub_receiver_stop_on_disconnect(create_redis, server, loop):
    pub = await create_redis(server.tcp_address, loop=loop)
    sub = await create_redis(server.tcp_address, loop=loop)
    sub_name = 'sub-{:X}'.format(id(sub))
    await sub.client_setname(sub_name)
    for sub_info in await pub.client_list():
        if sub_info.name == sub_name:
            break
    assert sub_info.name == sub_name

    mpsc = Receiver(loop=loop)
    await sub.subscribe(mpsc.channel('channel:1'))
    await sub.subscribe(mpsc.channel('channel:2'))
    await sub.psubscribe(mpsc.pattern('channel:*'))

    q = asyncio.Queue(loop=loop)
    EOF = object()

    async def reader():
        async for ch, msg in mpsc.iter(encoding='utf-8'):
            await q.put((ch.name, msg))
        await q.put(EOF)

    tsk = asyncio.ensure_future(reader(), loop=loop)
    await pub.publish_json('channel:1', ['hello'])
    await pub.publish_json('channel:2', ['hello'])
    # receive all messages
    assert await q.get() == (b'channel:1', '["hello"]')
    assert await q.get() == (b'channel:*', (b'channel:1', '["hello"]'))
    assert await q.get() == (b'channel:2', '["hello"]')
    assert await q.get() == (b'channel:*', (b'channel:2', '["hello"]'))

    # XXX: need to implement `client kill`
    assert await pub.execute('client', 'kill', sub_info.addr) in (b'OK', 1)
    await asyncio.wait_for(tsk, timeout=1, loop=loop)
    assert await q.get() is EOF



import asyncio
import pytest
import socket
import subprocess
import sys
import contextlib
import os
import ssl
import time
import logging
import tempfile
import atexit

from collections import namedtuple
from urllib.parse import urlencode, urlunparse
from async_timeout import timeout as async_timeout

import aioredis
import aioredis.sentinel


TCPAddress = namedtuple('TCPAddress', 'host port')

RedisServer = namedtuple('RedisServer',
                         'name tcp_address unixsocket version password')

SentinelServer = namedtuple('SentinelServer',
                            'name tcp_address unixsocket version masters')

# Public fixtures


@pytest.yield_fixture
def loop():
    """Creates new event loop."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)

    try:
        yield loop
    finally:
        if hasattr(loop, 'is_closed'):
            closed = loop.is_closed()
        else:
            closed = loop._closed   # XXX
        if not closed:
            loop.call_soon(loop.stop)
            loop.run_forever()
            loop.close()


@pytest.fixture(scope='session')
def unused_port():
    """Gets random free port."""
    def fun():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('127.0.0.1', 0))
            return s.getsockname()[1]
    return fun


@pytest.fixture
def create_connection(_closable, loop):
    """Wrapper around aioredis.create_connection."""

    async def f(*args, **kw):
        kw.setdefault('loop', loop)
        conn = await aioredis.create_connection(*args, **kw)
        _closable(conn)
        return conn
    return f


@pytest.fixture(params=[
    aioredis.create_redis,
    aioredis.create_redis_pool],
    ids=['single', 'pool'])
def create_redis(_closable, loop, request):
    """Wrapper around aioredis.create_redis."""
    factory = request.param

    async def f(*args, **kw):
        kw.setdefault('loop', loop)
        redis = await factory(*args, **kw)
        _closable(redis)
        return redis
    return f


@pytest.fixture
def create_pool(_closable, loop):
    """Wrapper around aioredis.create_pool."""

    async def f(*args, **kw):
        kw.setdefault('loop', loop)
        redis = await aioredis.create_pool(*args, **kw)
        _closable(redis)
        return redis
    return f


@pytest.fixture
def create_sentinel(_closable, loop):
    """Helper instantiating RedisSentinel client."""

    async def f(*args, **kw):
        kw.setdefault('loop', loop)
        # make it fail fast on slow CIs (if timeout argument is ommitted)
        kw.setdefault('timeout', .001)
        client = await aioredis.sentinel.create_sentinel(*args, **kw)
        _closable(client)
        return client
    return f


@pytest.fixture
def pool(create_pool, server, loop):
    """Returns RedisPool instance."""
    pool = loop.run_until_complete(
        create_pool(server.tcp_address, loop=loop))
    return pool


@pytest.fixture
def redis(create_redis, server, loop):
    """Returns Redis client instance."""
    redis = loop.run_until_complete(
        create_redis(server.tcp_address, loop=loop))
    loop.run_until_complete(redis.flushall())
    return redis


@pytest.fixture
def redis_sentinel(create_sentinel, sentinel, loop):
    """Returns Redis Sentinel client instance."""
    redis_sentinel = loop.run_until_complete(
        create_sentinel([sentinel.tcp_address], timeout=2, loop=loop))
    assert loop.run_until_complete(redis_sentinel.ping()) == b'PONG'
    return redis_sentinel


@pytest.yield_fixture
def _closable(loop):
    conns = []

    try:
        yield conns.append
    finally:
        waiters = []
        while conns:
            conn = conns.pop(0)
            conn.close()
            waiters.append(conn.wait_closed())
        if waiters:
            loop.run_until_complete(asyncio.gather(*waiters, loop=loop))


@pytest.fixture(scope='session')
def server(start_server):
    """Starts redis-server instance."""
    return start_server('A')


@pytest.fixture(scope='session')
def serverB(start_server):
    """Starts redis-server instance."""
    return start_server('B')


@pytest.fixture(scope='session')
def sentinel(start_sentinel, request, start_server):
    """Starts redis-sentinel instance with one master -- masterA."""
    # Adding master+slave for normal (no failover) tests:
    master_no_fail = start_server('master-no-fail')
    start_server('slave-no-fail', slaveof=master_no_fail)
    # Adding master+slave for failover test;
    masterA = start_server('masterA')
    start_server('slaveA', slaveof=masterA)
    return start_sentinel('main', masterA, master_no_fail)


@pytest.fixture(params=['path', 'query'])
def server_tcp_url(server, request):

    def make(**kwargs):
        netloc = '{0.host}:{0.port}'.format(server.tcp_address)
        path = ''
        if request.param == 'path':
            if 'password' in kwargs:
                netloc = ':{0}@{1.host}:{1.port}'.format(
                    kwargs.pop('password'), server.tcp_address)
            if 'db' in kwargs:
                path = '/{}'.format(kwargs.pop('db'))
        query = urlencode(kwargs)
        return urlunparse(('redis', netloc, path, '', query, ''))
    return make


@pytest.fixture
def server_unix_url(server):

    def make(**kwargs):
        query = urlencode(kwargs)
        return urlunparse(('unix', '', server.unixsocket, '', query, ''))
    return make

# Internal stuff #


def pytest_addoption(parser):
    parser.addoption('--redis-server', default=[],
                     action="append",
                     help="Path to redis-server executable,"
                          " defaults to `%(default)s`")
    parser.addoption('--ssl-cafile', default='tests/ssl/cafile.crt',
                     help="Path to testing SSL CA file")
    parser.addoption('--ssl-dhparam', default='tests/ssl/dhparam.pem',
                     help="Path to testing SSL DH params file")
    parser.addoption('--ssl-cert', default='tests/ssl/cert.pem',
                     help="Path to testing SSL CERT file")
    parser.addoption('--uvloop', default=False,
                     action='store_true',
                     help="Run tests with uvloop")


def _read_server_version(redis_bin):
    args = [redis_bin, '--version']
    with subprocess.Popen(args, stdout=subprocess.PIPE) as proc:
        version = proc.stdout.readline().decode('utf-8')
    for part in version.split():
        if part.startswith('v='):
            break
    else:
        raise RuntimeError(
            "No version info can be found in {}".format(version))
    return tuple(map(int, part[2:].split('.')))


@contextlib.contextmanager
def config_writer(path):
    with open(path, 'wt') as f:
        def write(*args):
            print(*args, file=f)
        yield write


REDIS_SERVERS = []
VERSIONS = {}


def format_version(srv):
    return 'redis_v{}'.format('.'.join(map(str, VERSIONS[srv])))


@pytest.fixture(scope='session')
def start_server(_proc, request, unused_port, server_bin):
    """Starts Redis server instance.

    Caches instances by name.
    ``name`` param -- instance alias
    ``config_lines`` -- optional list of config directives to put in config
        (if no config_lines passed -- no config will be generated,
         for backward compatibility).
    """

    version = _read_server_version(server_bin)
    verbose = request.config.getoption('-v') > 3

    servers = {}

    def timeout(t):
        end = time.time() + t
        while time.time() <= end:
            yield True
        raise RuntimeError("Redis startup timeout expired")

    def maker(name, config_lines=None, *, slaveof=None, password=None):
        assert slaveof is None or isinstance(slaveof, RedisServer), slaveof
        if name in servers:
            return servers[name]

        port = unused_port()
        tcp_address = TCPAddress('localhost', port)
        if sys.platform == 'win32':
            unixsocket = None
        else:
            unixsocket = '/tmp/aioredis.{}.sock'.format(port)
        dumpfile = 'dump-{}.rdb'.format(port)
        data_dir = tempfile.gettempdir()
        dumpfile_path = os.path.join(data_dir, dumpfile)
        stdout_file = os.path.join(data_dir, 'aioredis.{}.stdout'.format(port))
        tmp_files = [dumpfile_path, stdout_file]
        if config_lines:
            config = os.path.join(data_dir, 'aioredis.{}.conf'.format(port))
            with config_writer(config) as write:
                write('daemonize no')
                write('save ""')
                write('dir ', data_dir)
                write('dbfilename', dumpfile)
                write('port', port)
                if unixsocket:
                    write('unixsocket', unixsocket)
                    tmp_files.append(unixsocket)
                if password:
                    write('requirepass "{}"'.format(password))
                write('# extra config')
                for line in config_lines:
                    write(line)
                if slaveof is not None:
                    write("slaveof {0.tcp_address.host} {0.tcp_address.port}"
                          .format(slaveof))
                    if password:
                        write('masterauth "{}"'.format(password))
            args = [config]
            tmp_files.append(config)
        else:
            args = ['--daemonize', 'no',
                    '--save', '""',
                    '--dir', data_dir,
                    '--dbfilename', dumpfile,
                    '--port', str(port),
                    ]
            if unixsocket:
                args += [
                    '--unixsocket', unixsocket,
                    ]
            if password:
                args += [
                    '--requirepass "{}"'.format(password)
                    ]
            if slaveof is not None:
                args += [
                    '--slaveof',
                    str(slaveof.tcp_address.host),
                    str(slaveof.tcp_address.port),
                    ]
                if password:
                    args += [
                        '--masterauth "{}"'.format(password)
                    ]
        f = open(stdout_file, 'w')
        atexit.register(f.close)
        proc = _proc(server_bin, *args,
                     stdout=f,
                     stderr=subprocess.STDOUT,
                     _clear_tmp_files=tmp_files)
        with open(stdout_file, 'rt') as f:
            for _ in timeout(10):
                assert proc.poll() is None, (
                    "Process terminated", proc.returncode)
                log = f.readline()
                if log and verbose:
                    print(name, ":", log, end='')
                if 'The server is now ready to accept connections ' in log:
                    break
            if slaveof is not None:
                for _ in timeout(10):
                    log = f.readline()
                    if log and verbose:
                        print(name, ":", log, end='')
                    if 'sync: Finished with success' in log:
                        break
        info = RedisServer(name, tcp_address, unixsocket, version, password)
        servers.setdefault(name, info)
        return info

    return maker


@pytest.fixture(scope='session')
def start_sentinel(_proc, request, unused_port, server_bin):
    """Starts Redis Sentinel instances."""
    version = _read_server_version(server_bin)
    verbose = request.config.getoption('-v') > 3

    sentinels = {}

    def timeout(t):
        end = time.time() + t
        while time.time() <= end:
            yield True
        raise RuntimeError("Redis startup timeout expired")

    def maker(name, *masters, quorum=1, noslaves=False):
        key = (name,) + masters
        if key in sentinels:
            return sentinels[key]
        port = unused_port()
        tcp_address = TCPAddress('localhost', port)
        data_dir = tempfile.gettempdir()
        config = os.path.join(
            data_dir, 'aioredis-sentinel.{}.conf'.format(port))
        stdout_file = os.path.join(
            data_dir, 'aioredis-sentinel.{}.stdout'.format(port))
        tmp_files = [config, stdout_file]
        if sys.platform == 'win32':
            unixsocket = None
        else:
            unixsocket = os.path.join(
                data_dir, 'aioredis-sentinel.{}.sock'.format(port))
            tmp_files.append(unixsocket)

        with config_writer(config) as write:
            write('daemonize no')
            write('save ""')
            write('port', port)
            if unixsocket:
                write('unixsocket', unixsocket)
            write('loglevel debug')
            for master in masters:
                write('sentinel monitor', master.name,
                      '127.0.0.1', master.tcp_address.port, quorum)
                write('sentinel down-after-milliseconds', master.name, '3000')
                write('sentinel failover-timeout', master.name, '3000')
                write('sentinel auth-pass', master.name, master.password)

        f = open(stdout_file, 'w')
        atexit.register(f.close)
        proc = _proc(server_bin,
                     config,
                     '--sentinel',
                     stdout=f,
                     stderr=subprocess.STDOUT,
                     _clear_tmp_files=tmp_files)
        # XXX: wait sentinel see all masters and slaves;
        all_masters = {m.name for m in masters}
        if noslaves:
            all_slaves = {}
        else:
            all_slaves = {m.name for m in masters}
        with open(stdout_file, 'rt') as f:
            for _ in timeout(30):
                assert proc.poll() is None, (
                    "Process terminated", proc.returncode)
                log = f.readline()
                if log and verbose:
                    print(name, ":", log, end='')
                for m in masters:
                    if '# +monitor master {}'.format(m.name) in log:
                        all_masters.discard(m.name)
                    if '* +slave slave' in log and \
                            '@ {}'.format(m.name) in log:
                        all_slaves.discard(m.name)
                if not all_masters and not all_slaves:
                    break
            else:
                raise RuntimeError("Could not start Sentinel")

        masters = {m.name: m for m in masters}
        info = SentinelServer(name, tcp_address, unixsocket, version, masters)
        sentinels.setdefault(key, info)
        return info
    return maker


@pytest.fixture(scope='session')
def ssl_proxy(_proc, request, unused_port):
    by_port = {}

    cafile = os.path.abspath(request.config.getoption('--ssl-cafile'))
    certfile = os.path.abspath(request.config.getoption('--ssl-cert'))
    dhfile = os.path.abspath(request.config.getoption('--ssl-dhparam'))
    assert os.path.exists(cafile), \
        "Missing SSL CA file, run `make certificate` to generate new one"
    assert os.path.exists(certfile), \
        "Missing SSL CERT file, run `make certificate` to generate new one"
    assert os.path.exists(dhfile), \
        "Missing SSL DH params, run `make certificate` to generate new one"

    ssl_ctx = ssl.create_default_context(cafile=cafile)
    ssl_ctx.check_hostname = False
    ssl_ctx.verify_mode = ssl.CERT_NONE
    ssl_ctx.load_dh_params(dhfile)

    def sockat(unsecure_port):
        if unsecure_port in by_port:
            return by_port[unsecure_port]

        secure_port = unused_port()
        _proc('/usr/bin/socat',
              'openssl-listen:{port},'
              'dhparam={param},'
              'cert={cert},verify=0,fork'
              .format(port=secure_port, param=dhfile, cert=certfile),
              'tcp-connect:localhost:{}'
              .format(unsecure_port)
              )
        time.sleep(1)   # XXX
        by_port[unsecure_port] = secure_port, ssl_ctx
        return secure_port, ssl_ctx

    return sockat


@pytest.yield_fixture(scope='session')
def _proc():
    processes = []
    tmp_files = set()

    def run(*commandline, _clear_tmp_files=(), **kwargs):
        proc = subprocess.Popen(commandline, **kwargs)
        processes.append(proc)
        tmp_files.update(_clear_tmp_files)
        return proc

    try:
        yield run
    finally:
        while processes:
            proc = processes.pop(0)
            proc.terminate()
            proc.wait()
        for path in tmp_files:
            try:
                os.remove(path)
            except OSError:
                pass


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name):
        if not callable(obj):
            return
        item = pytest.Function(name, parent=collector)
        if item.get_closest_marker('run_loop') is not None:
            # TODO: re-wrap with asyncio.coroutine if not native coroutine
            return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    """
    Run asyncio marked test functions in an event loop instead of a normal
    function call.
    """
    marker = pyfuncitem.get_closest_marker('run_loop')
    if marker is not None:
        funcargs = pyfuncitem.funcargs
        loop = funcargs['loop']
        testargs = {arg: funcargs[arg]
                    for arg in pyfuncitem._fixtureinfo.argnames}

        loop.run_until_complete(
            _wait_coro(pyfuncitem.obj, testargs,
                       timeout=marker.kwargs.get('timeout', 15),
                       loop=loop))
        return True


async def _wait_coro(corofunc, kwargs, timeout, loop):
    with async_timeout(timeout, loop=loop):
        return (await corofunc(**kwargs))


def pytest_runtest_setup(item):
    run_loop = item.get_closest_marker('run_loop')
    if run_loop and 'loop' not in item.fixturenames:
        # inject an event loop fixture for all async tests
        item.fixturenames.append('loop')


def pytest_collection_modifyitems(session, config, items):
    skip_by_version = []
    for item in items[:]:
        marker = item.get_closest_marker('redis_version')
        if marker is not None:
            try:
                version = VERSIONS[item.callspec.getparam('server_bin')]
            except (KeyError, ValueError, AttributeError):
                # TODO: throw noisy warning
                continue
            if version < marker.kwargs['version']:
                skip_by_version.append(item)
                item.add_marker(pytest.mark.skip(
                    reason=marker.kwargs['reason']))
        if 'ssl_proxy' in item.fixturenames:
            item.add_marker(pytest.mark.skipif(
                "not os.path.exists('/usr/bin/socat')",
                reason="socat package required (apt-get install socat)"))
    if len(items) != len(skip_by_version):
        for i in skip_by_version:
            items.remove(i)


def pytest_configure(config):
    bins = config.getoption('--redis-server')[:]
    REDIS_SERVERS[:] = bins or ['/usr/bin/redis-server']
    VERSIONS.update({srv: _read_server_version(srv)
                     for srv in REDIS_SERVERS})
    assert VERSIONS, ("Expected to detect redis versions", REDIS_SERVERS)

    class DynamicFixturePlugin:
        @pytest.fixture(scope='session',
                        params=REDIS_SERVERS,
                        ids=format_version)
        def server_bin(self, request):
            """Common for start_server and start_sentinel
            server bin path parameter.
            """
            return request.param
    config.pluginmanager.register(DynamicFixturePlugin(), 'server-bin-fixture')

    if config.getoption('--uvloop'):
        try:
            import uvloop
        except ImportError:
            raise RuntimeError(
                "Can not import uvloop, make sure it is installed")
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


def logs(logger, level=None):
    """Catches logs for given logger and level.

    See unittest.TestCase.assertLogs for details.
    """
    return _AssertLogsContext(logger, level)


_LoggingWatcher = namedtuple("_LoggingWatcher", ["records", "output"])


class _CapturingHandler(logging.Handler):
    """
    A logging handler capturing all (raw and formatted) logging output.
    """

    def __init__(self):
        logging.Handler.__init__(self)
        self.watcher = _LoggingWatcher([], [])

    def flush(self):
        pass

    def emit(self, record):
        self.watcher.records.append(record)
        msg = self.format(record)
        self.watcher.output.append(msg)


class _AssertLogsContext:
    """Standard unittest's _AssertLogsContext context manager
    adopted to raise pytest failure.
    """
    LOGGING_FORMAT = "%(levelname)s:%(name)s:%(message)s"

    def __init__(self, logger_name, level):
        self.logger_name = logger_name
        if level:
            self.level = level
        else:
            self.level = logging.INFO
        self.msg = None

    def __enter__(self):
        if isinstance(self.logger_name, logging.Logger):
            logger = self.logger = self.logger_name
        else:
            logger = self.logger = logging.getLogger(self.logger_name)
        formatter = logging.Formatter(self.LOGGING_FORMAT)
        handler = _CapturingHandler()
        handler.setFormatter(formatter)
        self.watcher = handler.watcher
        self.old_handlers = logger.handlers[:]
        self.old_level = logger.level
        self.old_propagate = logger.propagate
        logger.handlers = [handler]
        logger.setLevel(self.level)
        logger.propagate = False
        return handler.watcher

    def __exit__(self, exc_type, exc_value, tb):
        self.logger.handlers = self.old_handlers
        self.logger.propagate = self.old_propagate
        self.logger.setLevel(self.old_level)
        if exc_type is not None:
            # let unexpected exceptions pass through
            return False
        if len(self.watcher.records) == 0:
            pytest.fail(
                "no logs of level {} or higher triggered on {}"
                .format(logging.getLevelName(self.level), self.logger.name))


def redis_version(*version, reason):
    assert 1 < len(version) <= 3, version
    assert all(isinstance(v, int) for v in version), version
    return pytest.mark.redis_version(version=version, reason=reason)


def assert_almost_equal(first, second, places=None, msg=None, delta=None):
    assert not (places is None and delta is None), \
        "Both places and delta are not set, please set one"
    if delta is not None:
        assert abs(first - second) <= delta
    else:
        assert round(abs(first - second), places) == 0


def pytest_namespace():
    return {
        'assert_almost_equal': assert_almost_equal,
        'redis_version': redis_version,
        'logs': logs,
        }



import pytest

from aioredis.util import parse_url


@pytest.mark.parametrize('url,expected_address,expected_options', [
    # redis scheme
    ('redis://', ('localhost', 6379), {}),
    ('redis://localhost:6379', ('localhost', 6379), {}),
    ('redis://localhost:6379/', ('localhost', 6379), {}),
    ('redis://localhost:6379/0', ('localhost', 6379), {'db': 0}),
    ('redis://localhost:6379/1', ('localhost', 6379), {'db': 1}),
    ('redis://localhost:6379?db=1', ('localhost', 6379), {'db': 1}),
    ('redis://localhost:6379/?db=1', ('localhost', 6379), {'db': 1}),
    ('redis://redis-host', ('redis-host', 6379), {}),
    ('redis://redis-host', ('redis-host', 6379), {}),
    ('redis://host:1234', ('host', 1234), {}),
    ('redis://user@localhost', ('localhost', 6379), {}),
    ('redis://:secret@localhost', ('localhost', 6379), {'password': 'secret'}),
    ('redis://user:secret@localhost',
     ('localhost', 6379), {'password': 'secret'}),
    ('redis://localhost?password=secret',
     ('localhost', 6379), {'password': 'secret'}),
    ('redis://localhost?encoding=utf-8',
     ('localhost', 6379), {'encoding': 'utf-8'}),
    ('redis://localhost?ssl=true',
     ('localhost', 6379), {'ssl': True}),
    ('redis://localhost?timeout=1.0',
     ('localhost', 6379), {'timeout': 1.0}),
    ('redis://localhost?timeout=10',
     ('localhost', 6379), {'timeout': 10.0}),
    # rediss scheme
    ('rediss://', ('localhost', 6379), {'ssl': True}),
    ('rediss://localhost:6379', ('localhost', 6379), {'ssl': True}),
    ('rediss://localhost:6379/', ('localhost', 6379), {'ssl': True}),
    ('rediss://localhost:6379/0', ('localhost', 6379), {'ssl': True, 'db': 0}),
    ('rediss://localhost:6379/1', ('localhost', 6379), {'ssl': True, 'db': 1}),
    ('rediss://localhost:6379?db=1',
     ('localhost', 6379), {'ssl': True, 'db': 1}),
    ('rediss://localhost:6379/?db=1',
     ('localhost', 6379), {'ssl': True, 'db': 1}),
    ('rediss://redis-host', ('redis-host', 6379), {'ssl': True}),
    ('rediss://redis-host', ('redis-host', 6379), {'ssl': True}),
    ('rediss://host:1234', ('host', 1234), {'ssl': True}),
    ('rediss://user@localhost', ('localhost', 6379), {'ssl': True}),
    ('rediss://:secret@localhost',
     ('localhost', 6379), {'ssl': True, 'password': 'secret'}),
    ('rediss://user:secret@localhost',
     ('localhost', 6379), {'ssl': True, 'password': 'secret'}),
    ('rediss://localhost?password=secret',
     ('localhost', 6379), {'ssl': True, 'password': 'secret'}),
    ('rediss://localhost?encoding=utf-8',
     ('localhost', 6379), {'ssl': True, 'encoding': 'utf-8'}),
    ('rediss://localhost?timeout=1.0',
     ('localhost', 6379), {'ssl': True, 'timeout': 1.0}),
    ('rediss://localhost?timeout=10',
     ('localhost', 6379), {'ssl': True, 'timeout': 10.0}),
    # unix scheme
    ('unix:///', '/', {}),
    ('unix:///redis.sock?db=12', '/redis.sock', {'db': 12}),
    ('unix:///redis.sock?encoding=utf-8',
     '/redis.sock', {'encoding': 'utf-8'}),
    ('unix:///redis.sock?ssl=true',
     '/redis.sock', {'ssl': True}),
    ('unix:///redis.sock?timeout=12',
     '/redis.sock', {'timeout': 12}),
    # no scheme
    ('/some/path/to/socket', '/some/path/to/socket', {}),
    ('/some/path/to/socket?db=1', '/some/path/to/socket?db=1', {}),
])
def test_good_url(url, expected_address, expected_options):
    address, options = parse_url(url)
    assert address == expected_address
    assert options == expected_options


@pytest.mark.parametrize('url,expected_error', [
    ('bad-scheme://localhost:6379/',
     ("Unsupported URI scheme", 'bad-scheme')),
    ('redis:///?db=1&db=2',
     ("Multiple parameters are not allowed", "db", "2")),
    ('redis:///?db=',
     ("Empty parameters are not allowed", "db", "")),
    ('redis:///?foo=',
     ("Empty parameters are not allowed", "foo", "")),
    ('unix://',
     ('Empty path is not allowed', 'unix://')),
    ('unix://host:123/',
     ('Netlocation is not allowed for unix scheme', 'host:123')),
    ('unix://user:pass@host:123/',
     ('Netlocation is not allowed for unix scheme', 'user:pass@host:123')),
    ('unix://user:pass@/',
     ('Netlocation is not allowed for unix scheme', 'user:pass@')),
    ('redis:///01',
     ('Expected integer without leading zeroes', '01')),
    ('rediss:///01',
     ('Expected integer without leading zeroes', '01')),
    ('redis:///?db=01',
     ('Expected integer without leading zeroes', '01')),
    ('rediss:///?db=01',
     ('Expected integer without leading zeroes', '01')),
    ('redis:///1?db=2',
     ('Single DB value expected, got path and query', 1, 2)),
    ('rediss:///1?db=2',
     ('Single DB value expected, got path and query', 1, 2)),
    ('redis://:passwd@localhost/?password=passwd',
     ('Single password value is expected, got in net location and query')),
    ('redis:///?ssl=1',
     ("Expected 'ssl' param to be 'true' or 'false' only", '1')),
    ('redis:///?ssl=True',
     ("Expected 'ssl' param to be 'true' or 'false' only", 'True')),
])
def test_url_assertions(url, expected_error):
    with pytest.raises(AssertionError) as exc_info:
        parse_url(url)
    assert exc_info.value.args == (expected_error,)


@pytest.mark.parametrize('url', [
    'redis:///bad-db-num',
    'redis:///0/1',
    'redis:///?db=bad-num',
    'redis:///?db=-1',
])
def test_db_num_assertions(url):
    with pytest.raises(AssertionError, match="Invalid decimal integer"):
        parse_url(url)



from aioredis.util import wait_convert, wait_ok, _NOTSET


class StringCommandsMixin:
    """String commands mixin.

    For commands details see: http://redis.io/commands/#string
    """

    SET_IF_NOT_EXIST = 'SET_IF_NOT_EXIST'   # NX
    SET_IF_EXIST = 'SET_IF_EXIST'           # XX

    def append(self, key, value):
        """Append a value to key."""
        return self.execute(b'APPEND', key, value)

    def bitcount(self, key, start=None, end=None):
        """Count set bits in a string.

        :raises TypeError: if only start or end specified.
        """
        if start is None and end is not None:
            raise TypeError("both start and stop must be specified")
        elif start is not None and end is None:
            raise TypeError("both start and stop must be specified")
        elif start is not None and end is not None:
            args = (start, end)
        else:
            args = ()
        return self.execute(b'BITCOUNT', key, *args)

    def bitfield(self):
        raise NotImplementedError()

    def bitop_and(self, dest, key, *keys):
        """Perform bitwise AND operations between strings."""
        return self.execute(b'BITOP', b'AND', dest, key, *keys)

    def bitop_or(self, dest, key, *keys):
        """Perform bitwise OR operations between strings."""
        return self.execute(b'BITOP', b'OR', dest, key, *keys)

    def bitop_xor(self, dest, key, *keys):
        """Perform bitwise XOR operations between strings."""
        return self.execute(b'BITOP', b'XOR', dest, key, *keys)

    def bitop_not(self, dest, key):
        """Perform bitwise NOT operations between strings."""
        return self.execute(b'BITOP', b'NOT', dest, key)

    def bitpos(self, key, bit, start=None, end=None):
        """Find first bit set or clear in a string.

        :raises ValueError: if bit is not 0 or 1
        """
        if bit not in (1, 0):
            raise ValueError("bit argument must be either 1 or 0")
        bytes_range = []
        if start is not None:
            bytes_range.append(start)
        if end is not None:
            if start is None:
                bytes_range = [0, end]
            else:
                bytes_range.append(end)
        return self.execute(b'BITPOS', key, bit, *bytes_range)

    def decr(self, key):
        """Decrement the integer value of a key by one."""
        return self.execute(b'DECR', key)

    def decrby(self, key, decrement):
        """Decrement the integer value of a key by the given number.

        :raises TypeError: if decrement is not int
        """
        if not isinstance(decrement, int):
            raise TypeError("decrement must be of type int")
        return self.execute(b'DECRBY', key, decrement)

    def get(self, key, *, encoding=_NOTSET):
        """Get the value of a key."""
        return self.execute(b'GET', key, encoding=encoding)

    def getbit(self, key, offset):
        """Returns the bit value at offset in the string value stored at key.

        :raises TypeError: if offset is not int
        :raises ValueError: if offset is less than 0
        """
        if not isinstance(offset, int):
            raise TypeError("offset argument must be int")
        if offset < 0:
            raise ValueError("offset must be greater equal 0")
        return self.execute(b'GETBIT', key, offset)

    def getrange(self, key, start, end, *, encoding=_NOTSET):
        """Get a substring of the string stored at a key.

        :raises TypeError: if start or end is not int
        """
        if not isinstance(start, int):
            raise TypeError("start argument must be int")
        if not isinstance(end, int):
            raise TypeError("end argument must be int")
        return self.execute(b'GETRANGE', key, start, end, encoding=encoding)

    def getset(self, key, value, *, encoding=_NOTSET):
        """Set the string value of a key and return its old value."""
        return self.execute(b'GETSET', key, value, encoding=encoding)

    def incr(self, key):
        """Increment the integer value of a key by one."""
        return self.execute(b'INCR', key)

    def incrby(self, key, increment):
        """Increment the integer value of a key by the given amount.

        :raises TypeError: if increment is not int
        """
        if not isinstance(increment, int):
            raise TypeError("increment must be of type int")
        return self.execute(b'INCRBY', key, increment)

    def incrbyfloat(self, key, increment):
        """Increment the float value of a key by the given amount.

        :raises TypeError: if increment is not int
        """
        if not isinstance(increment, float):
            raise TypeError("increment must be of type int")
        fut = self.execute(b'INCRBYFLOAT', key, increment)
        return wait_convert(fut, float)

    def mget(self, key, *keys, encoding=_NOTSET):
        """Get the values of all the given keys."""
        return self.execute(b'MGET', key, *keys, encoding=encoding)

    def mset(self, key, value, *pairs):
        """Set multiple keys to multiple values.

        :raises TypeError: if len of pairs is not event number
        """
        if len(pairs) % 2 != 0:
            raise TypeError("length of pairs must be even number")
        fut = self.execute(b'MSET', key, value, *pairs)
        return wait_ok(fut)

    def msetnx(self, key, value, *pairs):
        """Set multiple keys to multiple values,
        only if none of the keys exist.

        :raises TypeError: if len of pairs is not event number
        """
        if len(pairs) % 2 != 0:
            raise TypeError("length of pairs must be even number")
        return self.execute(b'MSETNX', key, value, *pairs)

    def psetex(self, key, milliseconds, value):
        """Set the value and expiration in milliseconds of a key.

        :raises TypeError: if milliseconds is not int
        """
        if not isinstance(milliseconds, int):
            raise TypeError("milliseconds argument must be int")
        fut = self.execute(b'PSETEX', key, milliseconds, value)
        return wait_ok(fut)

    def set(self, key, value, *, expire=0, pexpire=0, exist=None):
        """Set the string value of a key.

        :raises TypeError: if expire or pexpire is not int
        """
        if expire and not isinstance(expire, int):
            raise TypeError("expire argument must be int")
        if pexpire and not isinstance(pexpire, int):
            raise TypeError("pexpire argument must be int")

        args = []
        if expire:
            args[:] = [b'EX', expire]
        if pexpire:
            args[:] = [b'PX', pexpire]

        if exist is self.SET_IF_EXIST:
            args.append(b'XX')
        elif exist is self.SET_IF_NOT_EXIST:
            args.append(b'NX')
        fut = self.execute(b'SET', key, value, *args)
        return wait_ok(fut)

    def setbit(self, key, offset, value):
        """Sets or clears the bit at offset in the string value stored at key.

        :raises TypeError: if offset is not int
        :raises ValueError: if offset is less than 0 or value is not 0 or 1
        """
        if not isinstance(offset, int):
            raise TypeError("offset argument must be int")
        if offset < 0:
            raise ValueError("offset must be greater equal 0")
        if value not in (0, 1):
            raise ValueError("value argument must be either 1 or 0")
        return self.execute(b'SETBIT', key, offset, value)

    def setex(self, key, seconds, value):
        """Set the value and expiration of a key.

        If seconds is float it will be multiplied by 1000
        coerced to int and passed to `psetex` method.

        :raises TypeError: if seconds is neither int nor float
        """
        if isinstance(seconds, float):
            return self.psetex(key, int(seconds * 1000), value)
        if not isinstance(seconds, int):
            raise TypeError("milliseconds argument must be int")
        fut = self.execute(b'SETEX', key, seconds, value)
        return wait_ok(fut)

    def setnx(self, key, value):
        """Set the value of a key, only if the key does not exist."""
        fut = self.execute(b'SETNX', key, value)
        return wait_convert(fut, bool)

    def setrange(self, key, offset, value):
        """Overwrite part of a string at key starting at the specified offset.

        :raises TypeError: if offset is not int
        :raises ValueError: if offset less than 0
        """
        if not isinstance(offset, int):
            raise TypeError("offset argument must be int")
        if offset < 0:
            raise ValueError("offset must be greater equal 0")
        return self.execute(b'SETRANGE', key, offset, value)

    def strlen(self, key):
        """Get the length of the value stored in a key."""
        return self.execute(b'STRLEN', key)



import asyncio
import pytest
import async_timeout

from unittest.mock import patch

from aioredis import (
    ReplyError,
    PoolClosedError,
    ConnectionClosedError,
    ConnectionsPool,
    MaxClientsError,
    )


def _assert_defaults(pool):
    assert isinstance(pool, ConnectionsPool)
    assert pool.minsize == 1
    assert pool.maxsize == 10
    assert pool.size == 1
    assert pool.freesize == 1
    assert pool._close_waiter is None


def test_connect(pool):
    _assert_defaults(pool)


def test_global_loop(create_pool, loop, server):
    asyncio.set_event_loop(loop)

    pool = loop.run_until_complete(create_pool(
        server.tcp_address))
    _assert_defaults(pool)


@pytest.mark.run_loop
async def test_clear(pool):
    _assert_defaults(pool)

    await pool.clear()
    assert pool.freesize == 0


@pytest.mark.run_loop
@pytest.mark.parametrize('minsize', [None, -100, 0.0, 100])
async def test_minsize(minsize, create_pool, loop, server):

    with pytest.raises(AssertionError):
        await create_pool(
            server.tcp_address,
            minsize=minsize, maxsize=10, loop=loop)


@pytest.mark.run_loop
@pytest.mark.parametrize('maxsize', [None, -100, 0.0, 1])
async def test_maxsize(maxsize, create_pool, loop, server):

    with pytest.raises(AssertionError):
        await create_pool(
            server.tcp_address,
            minsize=2, maxsize=maxsize, loop=loop)


@pytest.mark.run_loop
async def test_create_connection_timeout(create_pool, loop, server):
    with patch.object(loop, 'create_connection') as\
            open_conn_mock:
        open_conn_mock.side_effect = lambda *a, **kw: asyncio.sleep(0.2,
                                                                    loop=loop)
        with pytest.raises(asyncio.TimeoutError):
            await create_pool(
                server.tcp_address, loop=loop,
                create_connection_timeout=0.1)


def test_no_yield_from(pool):
    with pytest.raises(RuntimeError):
        with pool:
            pass    # pragma: no cover


@pytest.mark.run_loop
async def test_simple_command(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=10, loop=loop)

    with (await pool) as conn:
        msg = await conn.execute('echo', 'hello')
        assert msg == b'hello'
        assert pool.size == 10
        assert pool.freesize == 9
    assert pool.size == 10
    assert pool.freesize == 10


@pytest.mark.run_loop
async def test_create_new(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=1, loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (await pool):
        assert pool.size == 1
        assert pool.freesize == 0

        with (await pool):
            assert pool.size == 2
            assert pool.freesize == 0

    assert pool.size == 2
    assert pool.freesize == 2


@pytest.mark.run_loop
async def test_create_constraints(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=1, maxsize=1, loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (await pool):
        assert pool.size == 1
        assert pool.freesize == 0

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(pool.acquire(),
                                   timeout=0.2,
                                   loop=loop)


@pytest.mark.run_loop
async def test_create_no_minsize(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=0, maxsize=1, loop=loop)
    assert pool.size == 0
    assert pool.freesize == 0

    with (await pool):
        assert pool.size == 1
        assert pool.freesize == 0

        with pytest.raises(asyncio.TimeoutError):
            await asyncio.wait_for(pool.acquire(),
                                   timeout=0.2,
                                   loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1


@pytest.mark.run_loop
async def test_create_pool_cls(create_pool, loop, server):

    class MyPool(ConnectionsPool):
        pass

    pool = await create_pool(
        server.tcp_address,
        loop=loop,
        pool_cls=MyPool)

    assert isinstance(pool, MyPool)


@pytest.mark.run_loop
async def test_create_pool_cls_invalid(create_pool, loop, server):
    with pytest.raises(AssertionError):
        await create_pool(
            server.tcp_address,
            loop=loop,
            pool_cls=type)


@pytest.mark.run_loop
async def test_release_closed(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=1, loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (await pool) as conn:
        conn.close()
        await conn.wait_closed()
    assert pool.size == 0
    assert pool.freesize == 0


@pytest.mark.run_loop
async def test_release_pending(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=1, loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with pytest.logs('aioredis', 'WARNING') as cm:
        with (await pool) as conn:
            try:
                await asyncio.wait_for(
                    conn.execute(
                        b'blpop',
                        b'somekey:not:exists',
                        b'0'),
                    0.1,
                    loop=loop)
            except asyncio.TimeoutError:
                pass
    assert pool.size == 0
    assert pool.freesize == 0
    assert cm.output == [
        'WARNING:aioredis:Connection <RedisConnection [db:0]>'
        ' has pending commands, closing it.'
    ]


@pytest.mark.run_loop
async def test_release_bad_connection(create_pool, create_redis, loop, server):
    pool = await create_pool(
        server.tcp_address,
        loop=loop)
    conn = await pool.acquire()
    assert conn.address[0] in ('127.0.0.1', '::1')
    assert conn.address[1] == server.tcp_address.port
    other_conn = await create_redis(
        server.tcp_address,
        loop=loop)
    with pytest.raises(AssertionError):
        pool.release(other_conn)

    pool.release(conn)
    other_conn.close()
    await other_conn.wait_closed()


@pytest.mark.run_loop
async def test_select_db(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        loop=loop)

    await pool.select(1)
    with (await pool) as conn:
        assert conn.db == 1


@pytest.mark.run_loop
async def test_change_db(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=1, db=0,
        loop=loop)
    assert pool.size == 1
    assert pool.freesize == 1

    with (await pool) as conn:
        await conn.select(1)
    assert pool.size == 0
    assert pool.freesize == 0

    with (await pool):
        assert pool.size == 1
        assert pool.freesize == 0

        await pool.select(1)
        assert pool.db == 1
        assert pool.size == 1
        assert pool.freesize == 0
    assert pool.size == 0
    assert pool.freesize == 0
    assert pool.db == 1


@pytest.mark.run_loop
async def test_change_db_errors(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        minsize=1, db=0,
        loop=loop)

    with pytest.raises(TypeError):
        await pool.select(None)
    assert pool.db == 0

    with (await pool):
        pass
    assert pool.size == 1
    assert pool.freesize == 1

    with pytest.raises(TypeError):
        await pool.select(None)
    assert pool.db == 0
    with pytest.raises(ValueError):
        await pool.select(-1)
    assert pool.db == 0
    with pytest.raises(ReplyError):
        await pool.select(100000)
    assert pool.db == 0


@pytest.mark.xfail(reason="Need to refactor this test")
@pytest.mark.run_loop
async def test_select_and_create(create_pool, loop, server):
    # trying to model situation when select and acquire
    # called simultaneously
    # but acquire freezes on _wait_select and
    # then continues with propper db

    # TODO: refactor this test as there's no _wait_select any more.
    with async_timeout.timeout(10, loop=loop):
        pool = await create_pool(
            server.tcp_address,
            minsize=1, db=0,
            loop=loop)
        db = 0
        while True:
            db = (db + 1) & 1
            _, conn = await asyncio.gather(pool.select(db),
                                           pool.acquire(),
                                           loop=loop)
            assert pool.db == db
            pool.release(conn)
            if conn.db == db:
                break
    # await asyncio.wait_for(test(), 3, loop=loop)


@pytest.mark.run_loop
async def test_response_decoding(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        encoding='utf-8', loop=loop)

    assert pool.encoding == 'utf-8'
    with (await pool) as conn:
        await conn.execute('set', 'key', 'value')
    with (await pool) as conn:
        res = await conn.execute('get', 'key')
        assert res == 'value'


@pytest.mark.run_loop
async def test_hgetall_response_decoding(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        encoding='utf-8', loop=loop)

    assert pool.encoding == 'utf-8'
    with (await pool) as conn:
        await conn.execute('del', 'key1')
        await conn.execute('hmset', 'key1', 'foo', 'bar')
        await conn.execute('hmset', 'key1', 'baz', 'zap')
    with (await pool) as conn:
        res = await conn.execute('hgetall', 'key1')
        assert res == ['foo', 'bar', 'baz', 'zap']


@pytest.mark.run_loop
async def test_crappy_multiexec(create_pool, loop, server):
    pool = await create_pool(
        server.tcp_address,
        encoding='utf-8', loop=loop,
        minsize=1, maxsize=1)

    with (await pool) as conn:
        await conn.execute('set', 'abc', 'def')
        await conn.execute('multi')
        await conn.execute('set', 'abc', 'fgh')
    assert conn.closed is True
    with (await pool) as conn:
        value = await conn.execute('get', 'abc')
    assert value == 'def'


@pytest.mark.run_loop
async def test_pool_size_growth(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address,
        loop=loop,
        minsize=1, maxsize=1)

    done = set()
    tasks = []

    async def task1(i):
        with (await pool):
            assert pool.size <= pool.maxsize
            assert pool.freesize == 0
            await asyncio.sleep(0.2, loop=loop)
            done.add(i)

    async def task2():
        with (await pool):
            assert pool.size <= pool.maxsize
            assert pool.freesize >= 0
            assert done == {0, 1}

    for _ in range(2):
        tasks.append(asyncio.ensure_future(task1(_), loop=loop))
    tasks.append(asyncio.ensure_future(task2(), loop=loop))
    await asyncio.gather(*tasks, loop=loop)


@pytest.mark.run_loop
async def test_pool_with_closed_connections(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address,
        loop=loop,
        minsize=1, maxsize=2)
    assert 1 == pool.freesize
    conn1 = pool._pool[0]
    conn1.close()
    assert conn1.closed is True
    assert 1 == pool.freesize
    with (await pool) as conn2:
        assert conn2.closed is False
        assert conn1 is not conn2


@pytest.mark.run_loop
async def test_pool_close(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address, loop=loop)

    assert pool.closed is False

    with (await pool) as conn:
        assert (await conn.execute('ping')) == b'PONG'

    pool.close()
    await pool.wait_closed()
    assert pool.closed is True

    with pytest.raises(PoolClosedError):
        with (await pool) as conn:
            assert (await conn.execute('ping')) == b'PONG'


@pytest.mark.run_loop
async def test_pool_close__used(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address, loop=loop)

    assert pool.closed is False

    with (await pool) as conn:
        pool.close()
        await pool.wait_closed()
        assert pool.closed is True

        with pytest.raises(ConnectionClosedError):
            await conn.execute('ping')


@pytest.mark.run_loop
@pytest.redis_version(2, 8, 0, reason="maxclients config setting")
async def test_pool_check_closed_when_exception(
        create_pool, create_redis, start_server, loop):
    server = start_server('server-small')
    redis = await create_redis(server.tcp_address, loop=loop)
    await redis.config_set('maxclients', 2)

    errors = (MaxClientsError, ConnectionClosedError, ConnectionError)
    with pytest.logs('aioredis', 'DEBUG') as cm:
        with pytest.raises(errors):
            await create_pool(address=tuple(server.tcp_address),
                              minsize=3, loop=loop)

    assert len(cm.output) >= 3
    connect_msg = (
        "DEBUG:aioredis:Creating tcp connection"
        " to ('localhost', {})".format(server.tcp_address.port))
    assert cm.output[:2] == [connect_msg, connect_msg]
    assert cm.output[-1] == "DEBUG:aioredis:Closed 1 connection(s)"


@pytest.mark.run_loop
async def test_pool_get_connection(create_pool, server, loop):
    pool = await create_pool(server.tcp_address, minsize=1, maxsize=2,
                             loop=loop)
    res = await pool.execute("set", "key", "val")
    assert res == b'OK'

    res = await pool.execute_pubsub("subscribe", "channel:1")
    assert res == [[b"subscribe", b"channel:1", 1]]

    res = await pool.execute("getset", "key", "value")
    assert res == b'val'

    res = await pool.execute_pubsub("subscribe", "channel:2")
    assert res == [[b"subscribe", b"channel:2", 2]]

    res = await pool.execute("get", "key")
    assert res == b'value'


@pytest.mark.run_loop
async def test_pool_get_connection_with_pipelining(create_pool, server, loop):
    pool = await create_pool(server.tcp_address, minsize=1, maxsize=2,
                             loop=loop)
    fut1 = pool.execute('set', 'key', 'val')
    fut2 = pool.execute_pubsub("subscribe", "channel:1")
    fut3 = pool.execute('getset', 'key', 'next')
    fut4 = pool.execute_pubsub("subscribe", "channel:2")
    fut5 = pool.execute('get', 'key')
    res = await fut1
    assert res == b'OK'
    res = await fut2
    assert res == [[b"subscribe", b"channel:1", 1]]
    res = await fut3
    assert res == b'val'
    res = await fut4
    assert res == [[b"subscribe", b"channel:2", 2]]
    res = await fut5
    assert res == b'next'


@pytest.mark.run_loop
async def test_pool_idle_close(create_pool, start_server, loop):
    server = start_server('idle')
    conn = await create_pool(server.tcp_address, minsize=2, loop=loop)
    ok = await conn.execute("config", "set", "timeout", 1)
    assert ok == b'OK'

    await asyncio.sleep(2, loop=loop)

    assert (await conn.execute('ping')) == b'PONG'


@pytest.mark.run_loop
async def test_await(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address,
        minsize=10, loop=loop)

    with (await pool) as conn:
        msg = await conn.execute('echo', 'hello')
        assert msg == b'hello'


@pytest.mark.run_loop
async def test_async_with(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address,
        minsize=10, loop=loop)

    async with pool.get() as conn:
        msg = await conn.execute('echo', 'hello')
        assert msg == b'hello'


@pytest.mark.run_loop
async def test_pool__drop_closed(create_pool, server, loop):
    pool = await create_pool(server.tcp_address,
                             minsize=3,
                             maxsize=3,
                             loop=loop)
    assert pool.size == 3
    assert pool.freesize == 3
    assert not pool._pool[0].closed
    assert not pool._pool[1].closed
    assert not pool._pool[2].closed

    pool._pool[1].close()
    pool._pool[2].close()
    await pool._pool[1].wait_closed()
    await pool._pool[2].wait_closed()

    assert not pool._pool[0].closed
    assert pool._pool[1].closed
    assert pool._pool[2].closed

    assert pool.size == 3
    assert pool.freesize == 3

    pool._drop_closed()
    assert pool.freesize == 1
    assert pool.size == 1



import asyncio
import aioredis


STOPWORD = 'STOP'


async def pubsub():
    pool = await aioredis.create_pool(
        'redis://localhost',
        minsize=5, maxsize=10)

    async def reader(channel):
        while (await channel.wait_message()):
            msg = await channel.get(encoding='utf-8')
            # ... process message ...
            print("message in {}: {}".format(channel.name, msg))

            if msg == STOPWORD:
                return

    with await pool as conn:
        await conn.execute_pubsub('subscribe', 'channel:1')
        channel = conn.pubsub_channels['channel:1']
        await reader(channel)  # wait for reader to complete
        await conn.execute_pubsub('unsubscribe', 'channel:1')

    # Explicit connection usage
    conn = await pool.acquire()
    try:
        await conn.execute_pubsub('subscribe', 'channel:1')
        channel = conn.pubsub_channels['channel:1']
        await reader(channel)  # wait for reader to complete
        await conn.execute_pubsub('unsubscribe', 'channel:1')
    finally:
        pool.release(conn)

    pool.close()
    await pool.wait_closed()    # closing all open connections


def main():
    loop = asyncio.get_event_loop()
    tsk = asyncio.ensure_future(pubsub(), loop=loop)

    async def publish():
        pub = await aioredis.create_redis(
            'redis://localhost')
        while not tsk.done():
            # wait for clients to subscribe
            while True:
                subs = await pub.pubsub_numsub('channel:1')
                if subs[b'channel:1'] == 1:
                    break
                await asyncio.sleep(0, loop=loop)
            # publish some messages
            for msg in ['one', 'two', 'three']:
                await pub.publish('channel:1', msg)
            # send stop word
            await pub.publish('channel:1', STOPWORD)
        pub.close()
        await pub.wait_closed()

    loop.run_until_complete(asyncio.gather(publish(), tsk, loop=loop))


if __name__ == '__main__':
    import os
    if 'redis_version:2.6' not in os.environ.get('REDIS_VERSION', ''):
        main()



from aioredis.util import wait_ok


class HyperLogLogCommandsMixin:
    """HyperLogLog commands mixin.

    For commands details see: http://redis.io/commands#hyperloglog
    """

    def pfadd(self, key, value, *values):
        """Adds the specified elements to the specified HyperLogLog."""
        return self.execute(b'PFADD', key, value, *values)

    def pfcount(self, key, *keys):
        """Return the approximated cardinality of
        the set(s) observed by the HyperLogLog at key(s).
        """
        return self.execute(b'PFCOUNT', key, *keys)

    def pfmerge(self, destkey, sourcekey, *sourcekeys):
        """Merge N different HyperLogLogs into a single one."""
        fut = self.execute(b'PFMERGE', destkey, sourcekey, *sourcekeys)
        return wait_ok(fut)



import asyncio
import functools

from ..abc import AbcPool
from ..errors import (
    RedisError,
    PipelineError,
    MultiExecError,
    ConnectionClosedError,
    )
from ..util import (
    wait_ok,
    _set_exception,
    )


class TransactionsCommandsMixin:
    """Transaction commands mixin.

    For commands details see: http://redis.io/commands/#transactions

    Transactions HOWTO:

    >>> tr = redis.multi_exec()
    >>> result_future1 = tr.incr('foo')
    >>> result_future2 = tr.incr('bar')
    >>> try:
    ...     result = await tr.execute()
    ... except MultiExecError:
    ...     pass    # check what happened
    >>> result1 = await result_future1
    >>> result2 = await result_future2
    >>> assert result == [result1, result2]
    """

    def unwatch(self):
        """Forget about all watched keys."""
        fut = self._pool_or_conn.execute(b'UNWATCH')
        return wait_ok(fut)

    def watch(self, key, *keys):
        """Watch the given keys to determine execution of the MULTI/EXEC block.
        """
        # FIXME: we can send watch through one connection and then issue
        #   'multi/exec' command through other.
        # Possible fix:
        #   "Remember" a connection that was used for 'watch' command
        #   and then send 'multi / exec / discard' through it.
        fut = self._pool_or_conn.execute(b'WATCH', key, *keys)
        return wait_ok(fut)

    def multi_exec(self):
        """Returns MULTI/EXEC pipeline wrapper.

        Usage:

        >>> tr = redis.multi_exec()
        >>> fut1 = tr.incr('foo')   # NO `await` as it will block forever!
        >>> fut2 = tr.incr('bar')
        >>> result = await tr.execute()
        >>> result
        [1, 1]
        >>> await asyncio.gather(fut1, fut2)
        [1, 1]
        """
        return MultiExec(self._pool_or_conn, self.__class__,
                         loop=self._pool_or_conn._loop)

    def pipeline(self):
        """Returns :class:`Pipeline` object to execute bulk of commands.

        It is provided for convenience.
        Commands can be pipelined without it.

        Example:

        >>> pipe = redis.pipeline()
        >>> fut1 = pipe.incr('foo') # NO `await` as it will block forever!
        >>> fut2 = pipe.incr('bar')
        >>> result = await pipe.execute()
        >>> result
        [1, 1]
        >>> await asyncio.gather(fut1, fut2)
        [1, 1]
        >>> #
        >>> # The same can be done without pipeline:
        >>> #
        >>> fut1 = redis.incr('foo')    # the 'INCRY foo' command already sent
        >>> fut2 = redis.incr('bar')
        >>> await asyncio.gather(fut1, fut2)
        [2, 2]
        """
        return Pipeline(self._pool_or_conn, self.__class__,
                        loop=self._pool_or_conn._loop)


class _RedisBuffer:

    def __init__(self, pipeline, *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._pipeline = pipeline
        self._loop = loop

    def execute(self, cmd, *args, **kw):
        fut = self._loop.create_future()
        self._pipeline.append((fut, cmd, args, kw))
        return fut

    # TODO: add here or remove in connection methods like `select`, `auth` etc


class Pipeline:
    """Commands pipeline.

    Usage:

    >>> pipe = redis.pipeline()
    >>> fut1 = pipe.incr('foo')
    >>> fut2 = pipe.incr('bar')
    >>> await pipe.execute()
    [1, 1]
    >>> await fut1
    1
    >>> await fut2
    1
    """
    error_class = PipelineError

    def __init__(self, pool_or_connection, commands_factory=lambda conn: conn,
                 *, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        self._pool_or_conn = pool_or_connection
        self._loop = loop
        self._pipeline = []
        self._results = []
        self._buffer = _RedisBuffer(self._pipeline, loop=loop)
        self._redis = commands_factory(self._buffer)
        self._done = False

    def __getattr__(self, name):
        assert not self._done, "Pipeline already executed. Create new one."
        attr = getattr(self._redis, name)
        if callable(attr):

            @functools.wraps(attr)
            def wrapper(*args, **kw):
                try:
                    task = asyncio.ensure_future(attr(*args, **kw),
                                                 loop=self._loop)
                except Exception as exc:
                    task = self._loop.create_future()
                    task.set_exception(exc)
                self._results.append(task)
                return task
            return wrapper
        return attr

    async def execute(self, *, return_exceptions=False):
        """Execute all buffered commands.

        Any exception that is raised by any command is caught and
        raised later when processing results.

        Exceptions can also be returned in result if
        `return_exceptions` flag is set to True.
        """
        assert not self._done, "Pipeline already executed. Create new one."
        self._done = True

        if self._pipeline:
            if isinstance(self._pool_or_conn, AbcPool):
                async with self._pool_or_conn.get() as conn:
                    return await self._do_execute(
                        conn, return_exceptions=return_exceptions)
            else:
                return await self._do_execute(
                    self._pool_or_conn,
                    return_exceptions=return_exceptions)
        else:
            return await self._gather_result(return_exceptions)

    async def _do_execute(self, conn, *, return_exceptions=False):
        await asyncio.gather(*self._send_pipeline(conn),
                             loop=self._loop,
                             return_exceptions=True)
        return await self._gather_result(return_exceptions)

    async def _gather_result(self, return_exceptions):
        errors = []
        results = []
        for fut in self._results:
            try:
                res = await fut
                results.append(res)
            except Exception as exc:
                errors.append(exc)
                results.append(exc)
        if errors and not return_exceptions:
            raise self.error_class(errors)
        return results

    def _send_pipeline(self, conn):
        with conn._buffered():
            for fut, cmd, args, kw in self._pipeline:
                try:
                    result_fut = conn.execute(cmd, *args, **kw)
                    result_fut.add_done_callback(
                        functools.partial(self._check_result, waiter=fut))
                except Exception as exc:
                    fut.set_exception(exc)
                else:
                    yield result_fut

    def _check_result(self, fut, waiter):
        if fut.cancelled():
            waiter.cancel()
        elif fut.exception():
            waiter.set_exception(fut.exception())
        else:
            waiter.set_result(fut.result())


class MultiExec(Pipeline):
    """Multi/Exec pipeline wrapper.

    Usage:

    >>> tr = redis.multi_exec()
    >>> f1 = tr.incr('foo')
    >>> f2 = tr.incr('bar')
    >>> # A)
    >>> await tr.execute()
    >>> res1 = await f1
    >>> res2 = await f2
    >>> # or B)
    >>> res1, res2 = await tr.execute()

    and ofcourse try/except:

    >>> tr = redis.multi_exec()
    >>> f1 = tr.incr('1') # won't raise any exception (why?)
    >>> try:
    ...     res = await tr.execute()
    ... except RedisError:
    ...     pass
    >>> assert f1.done()
    >>> assert f1.result() is res

    >>> tr = redis.multi_exec()
    >>> wait_ok_coro = tr.mset('1')
    >>> try:
    ...     ok1 = await tr.execute()
    ... except RedisError:
    ...     pass # handle it
    >>> ok2 = await wait_ok_coro
    >>> # for this to work `wait_ok_coro` must be wrapped in Future
    """
    error_class = MultiExecError

    async def _do_execute(self, conn, *, return_exceptions=False):
        self._waiters = waiters = []
        with conn._buffered():
            multi = conn.execute('MULTI')
            coros = list(self._send_pipeline(conn))
            exec_ = conn.execute('EXEC')
        gather = asyncio.gather(multi, *coros, loop=self._loop,
                                return_exceptions=True)
        last_error = None
        try:
            await asyncio.shield(gather, loop=self._loop)
        except asyncio.CancelledError:
            await gather
        except Exception as err:
            last_error = err
            raise
        finally:
            if conn.closed:
                if last_error is None:
                    last_error = ConnectionClosedError()
                for fut in waiters:
                    _set_exception(fut, last_error)
                    # fut.cancel()
                for fut in self._results:
                    if not fut.done():
                        fut.set_exception(last_error)
                        # fut.cancel()
            else:
                try:
                    results = await exec_
                except RedisError as err:
                    for fut in waiters:
                        fut.set_exception(err)
                else:
                    assert len(results) == len(waiters), (
                        "Results does not match waiters", results, waiters)
                    self._resolve_waiters(results, return_exceptions)
            return (await self._gather_result(return_exceptions))

    def _resolve_waiters(self, results, return_exceptions):
        errors = []
        for val, fut in zip(results, self._waiters):
            if isinstance(val, RedisError):
                fut.set_exception(val)
                errors.append(val)
            else:
                fut.set_result(val)
        if errors and not return_exceptions:
            raise MultiExecError(errors)

    def _check_result(self, fut, waiter):
        assert waiter not in self._waiters, (fut, waiter, self._waiters)
        assert not waiter.done(), waiter
        if fut.cancelled():     # await gather was cancelled
            waiter.cancel()
        elif fut.exception():   # server replied with error
            waiter.set_exception(fut.exception())
        elif fut.result() in {b'QUEUED', 'QUEUED'}:
            # got result, it should be QUEUED
            self._waiters.append(waiter)



import pytest

from aioredis.stream import StreamReader
from aioredis.parser import PyReader
from aioredis.errors import (
    ProtocolError,
    ReplyError
)


@pytest.fixture
def reader(loop):
    reader = StreamReader(loop=loop)
    reader.set_parser(
        PyReader(protocolError=ProtocolError, replyError=ReplyError)
    )
    return reader


@pytest.mark.run_loop
async def test_feed_and_parse(reader):
    reader.feed_data(b'+PONG\r\n')
    assert (await reader.readobj()) == b'PONG'


@pytest.mark.run_loop
async def test_buffer_available_after_RST(reader):
    reader.feed_data(b'+PONG\r\n')
    reader.set_exception(Exception())
    assert (await reader.readobj()) == b'PONG'
    with pytest.raises(Exception):
        await reader.readobj()


def test_feed_with_eof(reader):
    reader.feed_eof()
    with pytest.raises(AssertionError):
        reader.feed_data(b'+PONG\r\n')


def test_feed_no_data(reader):
    assert not reader.feed_data(None)


@pytest.mark.parametrize(
    'read_method',
    ['read', 'readline', 'readuntil', 'readexactly']
)
@pytest.mark.run_loop
async def test_read_flavors_not_supported(reader, read_method):
    with pytest.raises(RuntimeError):
        await getattr(reader, read_method)()



from .errors import ProtocolError, ReplyError

__all__ = [
    'Reader', 'PyReader',
]


class PyReader:
    """Pure-Python Redis protocol parser that follows hiredis.Reader
    interface (except setmaxbuf/getmaxbuf).
    """
    def __init__(self, protocolError=ProtocolError, replyError=ReplyError,
                 encoding=None):
        if not callable(protocolError):
            raise TypeError("Expected a callable")
        if not callable(replyError):
            raise TypeError("Expected a callable")
        self._parser = Parser(protocolError, replyError, encoding)

    def feed(self, data, o=0, l=-1):
        """Feed data to parser."""
        if l == -1:
            l = len(data) - o
        if o < 0 or l < 0:
            raise ValueError("negative input")
        if o + l > len(data):
            raise ValueError("input is larger than buffer size")
        self._parser.buf.extend(data[o:o+l])

    def gets(self):
        """Get parsed value or False otherwise.

        Error replies are return as replyError exceptions (not raised).
        Protocol errors are raised.
        """
        return self._parser.parse_one()

    def setmaxbuf(self, size):
        """No-op."""
        pass

    def getmaxbuf(self):
        """No-op."""
        return 0


class Parser:
    def __init__(self, protocolError, replyError, encoding):
        self.buf = bytearray()
        self.pos = 0
        self.protocolError = protocolError
        self.replyError = replyError
        self.encoding = encoding
        self._err = None
        self._gen = None

    def waitsome(self, size):
        # keep yielding false until at least `size` bytes added to buf.
        while len(self.buf) < self.pos+size:
            yield False

    def waitany(self):
        yield from self.waitsome(len(self.buf) + 1)

    def readone(self):
        if not self.buf[self.pos:1]:
            yield from self.waitany()
        val = self.buf[self.pos:1]
        self.pos += 1
        return val

    def readline(self, size=None):
        if size is not None:
            if len(self.buf) < size + 2 + self.pos:
                yield from self.waitsome(size + 2)
            offset = self.pos + size
            if self.buf[offset:offset+2] != b'\r\n':
                raise self.error("Expected b'\r\n'")
        else:
            offset = self.buf.find(b'\r\n', self.pos)
            while offset < 0:
                yield from self.waitany()
                offset = self.buf.find(b'\r\n', self.pos)
        val = self.buf[self.pos:offset]
        self.pos = 0
        del self.buf[:offset + 2]
        return val

    def readint(self):
        try:
            return int((yield from self.readline()))
        except ValueError as exc:
            raise self.error(exc)

    def error(self, msg):
        self._err = self.protocolError(msg)
        return self._err

    def parse(self, is_bulk=False):
        if self._err is not None:
            raise self._err
        ctl = yield from self.readone()
        if ctl == b'+':
            val = yield from self.readline()
            if self.encoding is not None:
                try:
                    return val.decode(self.encoding)
                except UnicodeDecodeError:
                    pass
            return bytes(val)
        elif ctl == b'-':
            val = yield from self.readline()
            return self.replyError(val.decode('utf-8'))
        elif ctl == b':':
            return (yield from self.readint())
        elif ctl == b'$':
            val = yield from self.readint()
            if val == -1:
                return None
            val = yield from self.readline(val)
            if self.encoding:
                try:
                    return val.decode(self.encoding)
                except UnicodeDecodeError:
                    pass
            return bytes(val)
        elif ctl == b'*':
            val = yield from self.readint()
            if val == -1:
                return None
            bulk_array = []
            error = None
            for _ in range(val):
                try:
                    bulk_array.append((yield from self.parse(is_bulk=True)))
                except LookupError as err:
                    if error is None:
                        error = err
            if error is not None:
                raise error
            return bulk_array
        else:
            raise self.error("Invalid first byte: {!r}".format(ctl))

    def parse_one(self):
        if self._gen is None:
            self._gen = self.parse()
        try:
            self._gen.send(None)
        except StopIteration as exc:
            self._gen = None
            return exc.value
        except Exception:
            self._gen = None
            raise
        else:
            return False


try:
    import hiredis
    Reader = hiredis.Reader
except ImportError:
    Reader = PyReader
